# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of the 'aea interact' subcommand."""

import codecs
import os
from pathlib import Path
from typing import Optional, TYPE_CHECKING, Type, Union

import click

from aea.cli.utils.decorators import check_aea_project
from aea.cli.utils.exceptions import InterruptInputException
from aea.common import Address
from aea.configurations.base import ConnectionConfig, PackageType, PublicId
from aea.configurations.constants import (
    CONNECTIONS,
    DEFAULT_AEA_CONFIG_FILE,
    DEFAULT_CONNECTION,
    DEFAULT_PROTOCOL,
    PROTOCOLS,
    SIGNING_PROTOCOL,
    STATE_UPDATE_PROTOCOL,
    VENDOR,
)
from aea.configurations.loader import ConfigLoader
from aea.connections.base import Connection
from aea.crypto.wallet import CryptoStore
from aea.identity.base import Identity
from aea.mail.base import Envelope, Message
from aea.multiplexer import InBox, Multiplexer, OutBox
from aea.protocols.base import Protocol
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.protocols.dialogue.base import Dialogues


if TYPE_CHECKING:  # pragma: nocover
    from packages.fetchai.connections.stub.connection import (  # noqa: F401
        DEFAULT_INPUT_FILE_NAME,
        DEFAULT_OUTPUT_FILE_NAME,
        StubConnection,
    )
    from packages.fetchai.protocols.default.dialogues import (  # noqa: F401
        DefaultDialogue,
        DefaultDialogues,
    )
    from packages.fetchai.protocols.default.message import DefaultMessage  # noqa: F401


@click.command()
@click.pass_context
@check_aea_project
def interact(click_context: click.core.Context):  # pylint: disable=unused-argument
    """Interact with a running AEA via the stub connection."""
    click.echo("Starting AEA interaction channel...")
    _run_interaction_channel()


def _load_packages(agent_identity: Identity):
    """Load packages in the current interpreter."""
    default_protocol_id = PublicId.from_str(DEFAULT_PROTOCOL)
    Protocol.from_dir(
        os.path.join(
            VENDOR, default_protocol_id.author, PROTOCOLS, default_protocol_id.name
        )
    )
    signing_protocol_id = PublicId.from_str(SIGNING_PROTOCOL)
    Protocol.from_dir(
        os.path.join(
            VENDOR, signing_protocol_id.author, PROTOCOLS, signing_protocol_id.name
        )
    )
    state_update_protocol_id = PublicId.from_str(STATE_UPDATE_PROTOCOL)
    Protocol.from_dir(
        os.path.join(
            VENDOR,
            state_update_protocol_id.author,
            PROTOCOLS,
            state_update_protocol_id.name,
        )
    )
    default_connection_id = PublicId.from_str(DEFAULT_CONNECTION)
    Connection.from_dir(
        os.path.join(
            VENDOR,
            default_connection_id.author,
            CONNECTIONS,
            default_connection_id.name,
        ),
        agent_identity,
        CryptoStore(),
    )


def _run_interaction_channel():
    loader = ConfigLoader.from_configuration_type(PackageType.AGENT)
    agent_configuration = loader.load(Path(DEFAULT_AEA_CONFIG_FILE).open())
    agent_name = agent_configuration.name

    identity_stub = Identity(agent_name + "_interact", "interact")
    _load_packages(identity_stub)

    # load agent configuration file
    from packages.fetchai.connections.stub.connection import (  # noqa: F811 # pylint: disable=import-outside-toplevel
        DEFAULT_INPUT_FILE_NAME,
        DEFAULT_OUTPUT_FILE_NAME,
        StubConnection,
    )
    from packages.fetchai.protocols.default.dialogues import (  # noqa: F811 # pylint: disable=import-outside-toplevel
        DefaultDialogue,
        DefaultDialogues,
    )
    from packages.fetchai.protocols.default.message import (  # noqa: F811 # pylint: disable=import-outside-toplevel
        DefaultMessage,
    )

    # load stub connection
    configuration = ConnectionConfig(
        input_file=DEFAULT_OUTPUT_FILE_NAME,
        output_file=DEFAULT_INPUT_FILE_NAME,
        connection_id=StubConnection.connection_id,
    )

    stub_connection = StubConnection(
        configuration=configuration, identity=identity_stub
    )
    multiplexer = Multiplexer([stub_connection])
    inbox = InBox(multiplexer)
    outbox = OutBox(multiplexer)

    def role_from_first_message(  # pylint: disable=unused-argument
        message: Message, receiver_address: Address
    ) -> BaseDialogue.Role:
        """Infer the role of the agent from an incoming/outgoing first message

        :param message: an incoming/outgoing first message
        :param receiver_address: the address of the receiving agent
        :return: The role of the agent
        """
        return DefaultDialogue.Role.AGENT

    dialogues = DefaultDialogues(identity_stub.name, role_from_first_message)

    try:
        multiplexer.connect()
        while True:  # pragma: no cover
            _process_envelopes(agent_name, inbox, outbox, dialogues, DefaultMessage)

    except KeyboardInterrupt:
        click.echo("Interaction interrupted!")
    except BaseException as e:  # pylint: disable=broad-except # pragma: no cover
        click.echo(e)
    finally:
        multiplexer.disconnect()


def _process_envelopes(
    agent_name: str,
    inbox: InBox,
    outbox: OutBox,
    dialogues: Dialogues,
    message_class: Type[Message],
) -> None:
    """
    Process envelopes.

    :param agent_name: name of an agent.
    :param inbox: an inbox object.
    :param outbox: an outbox object.
    :param dialogues: the dialogues object.
    :param message_class: the message class.

    :return: None.
    """
    envelope = _try_construct_envelope(agent_name, dialogues, message_class)
    if envelope is None:
        _check_for_incoming_envelope(inbox, message_class)
    else:
        outbox.put(envelope)
        click.echo(_construct_message("sending", envelope, message_class))


def _check_for_incoming_envelope(inbox: InBox, message_class: Type[Message]):
    if not inbox.empty():
        envelope = inbox.get_nowait()
        if envelope is None:
            raise ValueError("Could not recover envelope from inbox.")
        click.echo(_construct_message("received", envelope, message_class))
    else:
        click.echo("Received no new envelope!")


def _construct_message(
    action_name: str, envelope: Envelope, message_class: Type[Message]
):
    action_name = action_name.title()
    msg = (
        message_class.serializer.decode(envelope.message)
        if isinstance(envelope.message, bytes)
        else envelope.message
    )
    message = (
        "\n{} envelope:\nto: "
        "{}\nsender: {}\nprotocol_id: {}\nmessage: {}\n".format(
            action_name, envelope.to, envelope.sender, envelope.protocol_id, msg,
        )
    )
    return message


def _try_construct_envelope(
    agent_name: str, dialogues: Dialogues, message_class: Type[Message]
) -> Optional[Envelope]:
    """Try construct an envelope from user input."""
    envelope = None  # type: Optional[Envelope]
    try:
        performative_str = "bytes"
        performative = message_class.Performative(performative_str)
        click.echo(
            f"Provide message of protocol '{str(message_class.protocol_id)}' for performative {performative_str}:"
        )
        message_escaped = input()  # nosec
        message_escaped = message_escaped.strip()
        if message_escaped == "":
            raise InterruptInputException
        message_decoded = codecs.decode(
            message_escaped.encode("utf-8"), "unicode-escape"
        )
        message = message_decoded.encode("utf-8")  # type: Union[str, bytes]
        msg, _ = dialogues.create(
            counterparty=agent_name, performative=performative, content=message,
        )
        envelope = Envelope(
            to=msg.to, sender=msg.sender, protocol_id=msg.protocol_id, message=msg,
        )
    except InterruptInputException:
        click.echo("Interrupting input, checking inbox ...")
    except KeyboardInterrupt as e:
        raise e
    except BaseException as e:  # pylint: disable=broad-except # pragma: no cover
        click.echo(e)
    return envelope



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains the tests configuration utils."""
from unittest.mock import MagicMock

from aea.configurations.base import (
    AgentConfig,
    ComponentId,
    ComponentType,
    ConnectionConfig,
    PublicId,
    SkillConfig,
)
from aea.configurations.utils import (
    get_latest_component_id_from_prefix,
    replace_component_ids,
)


def test_get_latest_component_id_from_prefix():
    """Test the utility to get the latest concrete version id."""
    agent_config = MagicMock()
    expected_component_id = ComponentId(
        ComponentType.PROTOCOL, PublicId("author", "name", "0.1.0")
    )
    agent_config.package_dependencies = {expected_component_id}

    result = get_latest_component_id_from_prefix(
        agent_config, expected_component_id.component_prefix
    )
    assert result == expected_component_id


def test_get_latest_component_id_from_prefix_negative():
    """Test the utility to get the latest concrete version id, negative case."""
    agent_config = MagicMock()
    agent_config.package_dependencies = {}

    result = get_latest_component_id_from_prefix(
        agent_config, (ComponentType.PROTOCOL, "author", "name")
    )
    assert result is None


class BaseTestReplaceComponentIds:
    """Base test class for 'replace_component_ids' utility function."""

    old_protocol_id = PublicId("author", "old_protocol", "0.1.0")
    old_contract_id = PublicId("author", "old_contract", "0.1.0")
    old_connection_id = PublicId("author", "old_connection", "0.1.0")
    old_skill_id = PublicId("author", "old_skill", "0.1.0")

    new_protocol_id = PublicId("author", "new_protocol", "0.1.0")
    new_contract_id = PublicId("author", "new_contract", "0.1.0")
    new_connection_id = PublicId("author", "new_connection", "0.1.0")
    new_skill_id = PublicId("author", "new_skill", "0.1.0")

    new_public_ids = {
        new_protocol_id,
        new_contract_id,
        new_connection_id,
        new_skill_id,
    }

    replacements = {
        ComponentType.PROTOCOL: {old_protocol_id: new_protocol_id},
        ComponentType.CONTRACT: {old_contract_id: new_contract_id},
        ComponentType.CONNECTION: {old_connection_id: new_connection_id},
        ComponentType.SKILL: {old_skill_id: new_skill_id},
    }


class TestReplaceComponentIdsInAgentConfig(BaseTestReplaceComponentIds):
    """Test replace component ids in agent configuration."""

    @classmethod
    def setup_class(cls):
        """Set up the test."""
        cls.expected_custom_component_configuration = dict(foo="bar")

        cls.agent_config = AgentConfig(
            agent_name="agent_name",
            author="author",
            version="0.1.0",
            default_routing={str(cls.old_protocol_id): str(cls.old_connection_id)},
            default_connection=str(cls.old_connection_id),
        )

        cls.agent_config.protocols = {cls.old_protocol_id}
        cls.agent_config.contracts = {cls.old_contract_id}
        cls.agent_config.connections = {cls.old_connection_id}
        cls.agent_config.skills = {cls.old_skill_id}
        cls.agent_config.component_configurations[
            ComponentId(ComponentType.PROTOCOL, cls.old_protocol_id)
        ] = cls.expected_custom_component_configuration
        cls.agent_config.component_configurations[
            ComponentId(ComponentType.CONTRACT, cls.old_contract_id)
        ] = cls.expected_custom_component_configuration
        cls.agent_config.component_configurations[
            ComponentId(ComponentType.CONNECTION, cls.old_connection_id)
        ] = cls.expected_custom_component_configuration
        cls.agent_config.component_configurations[
            ComponentId(ComponentType.SKILL, cls.old_skill_id)
        ] = cls.expected_custom_component_configuration

        replace_component_ids(cls.agent_config, cls.replacements)

    def test_protocols_updated(self):
        """Test set of protocol ids updated."""
        assert self.agent_config.protocols == {self.new_protocol_id}

    def test_contracts_updated(self):
        """Test set of contract ids updated."""
        assert self.agent_config.contracts == {self.new_contract_id}

    def test_connections_updated(self):
        """Test set of connection ids updated."""
        assert self.agent_config.connections == {self.new_connection_id}

    def test_skills_updated(self):
        """Test set of skill ids updated."""
        assert self.agent_config.skills == {self.new_skill_id}

    def test_default_connection_updated(self):
        """Test default connection updated."""
        assert self.agent_config.default_connection == self.new_connection_id

    def test_default_routing_updated(self):
        """Test default routing updated."""
        assert self.agent_config.default_routing == {
            self.new_protocol_id: self.new_connection_id
        }

    def test_custom_configuration_updated(self):
        """Test default routing updated."""
        component_protocol_id = ComponentId(
            ComponentType.PROTOCOL, self.new_protocol_id
        )
        component_contract_id = ComponentId(
            ComponentType.CONTRACT, self.new_contract_id
        )
        component_connection_id = ComponentId(
            ComponentType.CONNECTION, self.new_connection_id
        )
        component_skill_id = ComponentId(ComponentType.SKILL, self.new_skill_id)

        assert (
            self.agent_config.component_configurations[component_protocol_id]
            == self.expected_custom_component_configuration
        )
        assert (
            self.agent_config.component_configurations[component_contract_id]
            == self.expected_custom_component_configuration
        )
        assert (
            self.agent_config.component_configurations[component_connection_id]
            == self.expected_custom_component_configuration
        )
        assert (
            self.agent_config.component_configurations[component_skill_id]
            == self.expected_custom_component_configuration
        )


class TestReplaceComponentIdsInConnectionConfig(BaseTestReplaceComponentIds):
    """Test replace component ids utility with connection configuration."""

    @classmethod
    def setup_class(cls):
        """Set up the test."""
        cls.connection_config = ConnectionConfig(
            name="connection_name",
            author="author",
            version="0.1.0",
            connections={cls.old_connection_id},
            protocols={cls.old_protocol_id},
            restricted_to_protocols={cls.old_protocol_id},
            excluded_protocols={cls.old_protocol_id},
        )
        replace_component_ids(cls.connection_config, cls.replacements)

    def test_protocols_updated(self):
        """Test set of protocol ids updated."""
        assert self.connection_config.protocols == {self.new_protocol_id}

    def test_connections_updated(self):
        """Test set of connection ids updated."""
        assert self.connection_config.connections == {self.new_connection_id}

    def test_restricted_to_protocols_updated(self):
        """Test restricted to protocols updated."""
        assert self.connection_config.restricted_to_protocols == {self.new_protocol_id}

    def test_excluded_protocols_updated(self):
        """Test excluded protocols updated."""
        assert self.connection_config.excluded_protocols == {self.new_protocol_id}


class TestReplaceComponentIdsInSkillConfig(BaseTestReplaceComponentIds):
    """Test replace component ids in skill configuration."""

    @classmethod
    def setup_class(cls):
        """Set up the test class."""
        cls.expected_custom_component_configuration = dict(foo="bar")

        cls.skill_config = SkillConfig(
            name="skill_name", author="author", version="0.1.0",
        )

        cls.skill_config.protocols = {cls.old_protocol_id}
        cls.skill_config.contracts = {cls.old_contract_id}
        cls.skill_config.connections = {cls.old_connection_id}
        cls.skill_config.skills = {cls.old_skill_id}

        replace_component_ids(cls.skill_config, cls.replacements)

    def test_protocols_updated(self):
        """Test set of protocol ids updated."""
        assert self.skill_config.protocols == {self.new_protocol_id}

    def test_contracts_updated(self):
        """Test set of contract ids updated."""
        assert self.skill_config.contracts == {self.new_contract_id}

    def test_connections_updated(self):
        """Test set of connection ids updated."""
        assert self.skill_config.connections == {self.new_connection_id}

    def test_skills_updated(self):
        """Test set of skill ids updated."""
        assert self.skill_config.skills == {self.new_skill_id}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the strategy class."""

import datetime
from typing import Dict, List, Optional, Tuple, cast

from packages.fetchai.skills.confirmation_aw2.registration_db import RegistrationDB
from packages.fetchai.skills.generic_buyer.strategy import GenericStrategy


class Strategy(GenericStrategy):
    """Strategy class extending Generic Strategy."""

    def __init__(self, **kwargs) -> None:
        """
        Initialize the strategy of the agent.

        :return: None
        """
        aw1_aea: Optional[str] = kwargs.pop("aw1_aea", None)
        if aw1_aea is None:
            raise ValueError("aw1_aea must be provided!")
        self.aw1_aea = aw1_aea
        self.mininum_hours_between_txs = kwargs.pop("mininum_hours_between_txs", 4)
        self.minimum_minutes_since_last_attempt = kwargs.pop(
            "minimum_minutes_since_last_attempt", 2
        )
        super().__init__(**kwargs)
        self.last_attempt: Dict[str, datetime.datetime] = {}

    def get_acceptable_counterparties(
        self, counterparties: Tuple[str, ...]
    ) -> Tuple[str, ...]:
        """
        Process counterparties and drop unacceptable ones.

        :return: list of counterparties
        """
        valid_counterparties: List[str] = []
        for counterparty in counterparties:
            if self.is_valid_counterparty(counterparty):
                valid_counterparties.append(counterparty)
        return tuple(valid_counterparties)

    def is_enough_time_since_last_attempt(self, counterparty: str) -> bool:
        """
        Check if enough time has passed since last attempt for potential previous trade to complete.

        :return: bool indicating validity
        """
        last_time = self.last_attempt.get(counterparty, None)
        if last_time is None:
            return True
        result = datetime.datetime.now() > last_time + datetime.timedelta(
            minutes=self.minimum_minutes_since_last_attempt
        )
        return result

    def is_valid_counterparty(self, counterparty: str) -> bool:
        """
        Check if the counterparty is valid.

        :return: bool indicating validity
        """
        registration_db = cast(RegistrationDB, self.context.registration_db)
        if not registration_db.is_registered(counterparty):
            self.context.logger.info(
                f"Invalid counterparty={counterparty}, not registered!"
            )
            return False
        if not self.is_enough_time_since_last_attempt(counterparty):
            self.context.logger.debug(
                f"Not enough time since last attempt for counterparty={counterparty}!"
            )
            return False
        self.last_attempt[counterparty] = datetime.datetime.now()
        if not registration_db.is_allowed_to_trade(
            counterparty, self.mininum_hours_between_txs
        ):
            return False
        return True

    def successful_trade_with_counterparty(
        self, counterparty: str, data: Dict[str, str]
    ) -> None:
        """
        Do something on successful trade.

        :param counterparty: the counterparty address
        :param data: the data
        :return: False
        """
        registration_db = cast(RegistrationDB, self.context.registration_db)
        registration_db.set_trade(counterparty, datetime.datetime.now(), data)
        self.context.logger.info(
            f"Successful trade with={counterparty}. Data acquired={data}!"
        )

    def register_counterparty(self, counterparty: str) -> None:
        """
        Register a counterparty.

        :param counterparty: the counterparty address
        :param data: the data
        :return: False
        """
        registration_db = cast(RegistrationDB, self.context.registration_db)
        registration_db.set_registered(counterparty)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the decision maker class."""

import copy
import logging
from enum import Enum
from typing import Any, Dict, List, Optional, TYPE_CHECKING, cast

from aea.common import Address
from aea.crypto.wallet import Wallet
from aea.decision_maker.base import DecisionMakerHandler as BaseDecisionMakerHandler
from aea.decision_maker.base import OwnershipState as BaseOwnershipState
from aea.decision_maker.base import Preferences as BasePreferences
from aea.exceptions import enforce
from aea.helpers.preference_representations.base import (
    linear_utility,
    logarithmic_utility,
)
from aea.helpers.transaction.base import SignedMessage, SignedTransaction, Terms
from aea.identity.base import Identity
from aea.protocols.base import Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue


if TYPE_CHECKING:  # pragma: nocover
    from packages.fetchai.protocols.signing.dialogues import (  # noqa: F401
        SigningDialogue,
    )
    from packages.fetchai.protocols.signing.dialogues import (  # noqa: F401
        SigningDialogues as BaseSigningDialogues,
    )
    from packages.fetchai.protocols.signing.message import SigningMessage  # noqa: F401
    from packages.fetchai.protocols.state_update.dialogues import (  # noqa: F401
        StateUpdateDialogue,
    )
    from packages.fetchai.protocols.state_update.dialogues import (  # noqa: F401
        StateUpdateDialogues as BaseStateUpdateDialogues,
    )
    from packages.fetchai.protocols.state_update.message import (  # noqa: F401
        StateUpdateMessage,
    )


CurrencyHoldings = Dict[str, int]  # a map from identifier to quantity
GoodHoldings = Dict[str, int]  # a map from identifier to quantity
UtilityParams = Dict[str, float]  # a map from identifier to quantity
ExchangeParams = Dict[str, float]  # a map from identifier to quantity

QUANTITY_SHIFT = 100

_default_logger = logging.getLogger(__name__)


class GoalPursuitReadiness:
    """The goal pursuit readiness."""

    class Status(Enum):
        """
        The enum of the readiness status.

        In particular, it can be one of the following:

        - Status.READY: when the agent is ready to pursuit its goal
        - Status.NOT_READY: when the agent is not ready to pursuit its goal
        """

        READY = "ready"
        NOT_READY = "not_ready"

    def __init__(self):
        """Instantiate the goal pursuit readiness."""
        self._status = GoalPursuitReadiness.Status.NOT_READY

    @property
    def is_ready(self) -> bool:
        """Get the readiness."""
        return self._status.value == GoalPursuitReadiness.Status.READY.value

    def update(self, new_status: Status) -> None:
        """
        Update the goal pursuit readiness.

        :param new_status: the new status
        :return: None
        """
        self._status = new_status


class OwnershipState(BaseOwnershipState):
    """Represent the ownership state of an agent (can proxy a ledger)."""

    def __init__(self):
        """
        Instantiate an ownership state object.

        :param decision_maker: the decision maker
        """
        self._amount_by_currency_id = None  # type: Optional[CurrencyHoldings]
        self._quantities_by_good_id = None  # type: Optional[GoodHoldings]

    def set(  # pylint: disable=arguments-differ
        self,
        amount_by_currency_id: CurrencyHoldings = None,
        quantities_by_good_id: GoodHoldings = None,
        **kwargs,
    ) -> None:
        """
        Set values on the ownership state.

        :param amount_by_currency_id: the currency endowment of the agent in this state.
        :param quantities_by_good_id: the good endowment of the agent in this state.
        """
        if amount_by_currency_id is None:  # pragma: nocover
            raise ValueError("Must provide amount_by_currency_id.")
        if quantities_by_good_id is None:  # pragma: nocover
            raise ValueError("Must provide quantities_by_good_id.")
        enforce(
            not self.is_initialized,
            "Cannot apply state update, current state is already initialized!",
        )

        self._amount_by_currency_id = copy.copy(amount_by_currency_id)
        self._quantities_by_good_id = copy.copy(quantities_by_good_id)

    def apply_delta(  # pylint: disable=arguments-differ
        self,
        delta_amount_by_currency_id: Dict[str, int] = None,
        delta_quantities_by_good_id: Dict[str, int] = None,
        **kwargs,
    ) -> None:
        """
        Apply a state update to the ownership state.

        This method is used to apply a raw state update without a transaction.

        :param delta_amount_by_currency_id: the delta in the currency amounts
        :param delta_quantities_by_good_id: the delta in the quantities by good
        :return: None
        """
        if delta_amount_by_currency_id is None:  # pragma: nocover
            raise ValueError("Must provide delta_amount_by_currency_id.")
        if delta_quantities_by_good_id is None:  # pragma: nocover
            raise ValueError("Must provide delta_quantities_by_good_id.")
        if self._amount_by_currency_id is None or self._quantities_by_good_id is None:
            raise ValueError(  # pragma: nocover
                "Cannot apply state update, current state is not initialized!"
            )
        enforce(
            all(
                [
                    key in self._amount_by_currency_id
                    for key in delta_amount_by_currency_id.keys()
                ]
            ),
            "Invalid keys present in delta_amount_by_currency_id.",
        )
        enforce(
            all(
                [
                    key in self._quantities_by_good_id
                    for key in delta_quantities_by_good_id.keys()
                ]
            ),
            "Invalid keys present in delta_quantities_by_good_id.",
        )

        for currency_id, amount_delta in delta_amount_by_currency_id.items():
            self._amount_by_currency_id[currency_id] += amount_delta

        for good_id, quantity_delta in delta_quantities_by_good_id.items():
            self._quantities_by_good_id[good_id] += quantity_delta

    @property
    def is_initialized(self) -> bool:
        """Get the initialization status."""
        return (
            self._amount_by_currency_id is not None
            and self._quantities_by_good_id is not None
        )

    @property
    def amount_by_currency_id(self) -> CurrencyHoldings:
        """Get currency holdings in this state."""
        if self._amount_by_currency_id is None:
            raise ValueError("amount_by_currency_id is not set!")
        return copy.copy(self._amount_by_currency_id)

    @property
    def quantities_by_good_id(self) -> GoodHoldings:
        """Get good holdings in this state."""
        if self._quantities_by_good_id is None:
            raise ValueError("quantities_by_good_id is not set!")
        return copy.copy(self._quantities_by_good_id)

    def is_affordable_transaction(self, terms: Terms) -> bool:
        """
        Check if the transaction is affordable (and consistent).

        E.g. check that the agent state has enough money if it is a buyer or enough holdings if it is a seller.
        Note, the agent is the sender of the transaction message by design.

        :param terms: the transaction terms
        :return: True if the transaction is legal wrt the current state, false otherwise.
        """
        if all(amount == 0 for amount in terms.amount_by_currency_id.values()) and all(
            quantity == 0 for quantity in terms.quantities_by_good_id.values()
        ):
            # reject the transaction when there is no wealth exchange
            result = False
        elif all(
            amount <= 0 for amount in terms.amount_by_currency_id.values()
        ) and all(quantity >= 0 for quantity in terms.quantities_by_good_id.values()):
            # check if the agent has the money to cover the sender_amount (the agent=sender is the buyer)
            result = all(
                self.amount_by_currency_id[currency_id] >= -amount
                for currency_id, amount in terms.amount_by_currency_id.items()
            )
        elif all(
            amount >= 0 for amount in terms.amount_by_currency_id.values()
        ) and all(quantity <= 0 for quantity in terms.quantities_by_good_id.values()):
            # check if the agent has the goods (the agent=sender is the seller).
            result = all(
                self.quantities_by_good_id[good_id] >= -quantity
                for good_id, quantity in terms.quantities_by_good_id.items()
            )
        else:
            result = False
        return result

    def is_affordable(self, terms: Terms) -> bool:
        """
        Check if the tx is affordable.

        :param terms: the transaction terms
        :return: whether the transaction is affordable or not
        """
        if self.is_initialized:
            is_affordable = self.is_affordable_transaction(terms)
        else:
            _default_logger.debug(
                "Cannot verify whether transaction is affordable as ownership state is not initialized. Assuming it is!"
            )
            is_affordable = True
        return is_affordable

    def update(self, terms: Terms) -> None:
        """
        Update the agent state from a transaction.

        :param terms: the transaction terms
        :return: None
        """
        if self._amount_by_currency_id is None or self._quantities_by_good_id is None:
            raise ValueError(  # pragma: nocover
                "Cannot apply state update, current state is not initialized!"
            )
        for currency_id, amount_delta in terms.amount_by_currency_id.items():
            self._amount_by_currency_id[currency_id] += amount_delta

        for good_id, quantity_delta in terms.quantities_by_good_id.items():
            self._quantities_by_good_id[good_id] += quantity_delta

    def apply_transactions(self, list_of_terms: List[Terms]) -> "OwnershipState":
        """
        Apply a list of transactions to (a copy of) the current state.

        :param list_of_terms: the sequence of transaction terms.
        :return: the final state.
        """
        new_state = copy.copy(self)
        for terms in list_of_terms:
            new_state.update(terms)

        return new_state

    def __copy__(self) -> "OwnershipState":
        """Copy the object."""
        state = OwnershipState()
        if self.is_initialized:
            state._amount_by_currency_id = self.amount_by_currency_id
            state._quantities_by_good_id = self.quantities_by_good_id
        return state


class Preferences(BasePreferences):
    """Class to represent the preferences."""

    def __init__(self):
        """Instantiate an agent preference object."""
        self._exchange_params_by_currency_id = None  # type: Optional[ExchangeParams]
        self._utility_params_by_good_id = None  # type: Optional[UtilityParams]
        self._quantity_shift = QUANTITY_SHIFT

    def set(  # pylint: disable=arguments-differ
        self,
        exchange_params_by_currency_id: ExchangeParams = None,
        utility_params_by_good_id: UtilityParams = None,
        **kwargs,
    ) -> None:
        """
        Set values on the preferences.

        :param exchange_params_by_currency_id: the exchange params.
        :param utility_params_by_good_id: the utility params for every asset.
        """
        if exchange_params_by_currency_id is None:  # pragma: nocover
            raise ValueError("Must provide exchange_params_by_currency_id.")
        if utility_params_by_good_id is None:  # pragma: nocover
            raise ValueError("Must provide utility_params_by_good_id.")
        enforce(
            not self.is_initialized,
            "Cannot apply preferences update, preferences already initialized!",
        )

        self._exchange_params_by_currency_id = copy.copy(exchange_params_by_currency_id)
        self._utility_params_by_good_id = copy.copy(utility_params_by_good_id)

    @property
    def is_initialized(self) -> bool:
        """
        Get the initialization status.

        Returns True if exchange_params_by_currency_id and utility_params_by_good_id are not None.
        """
        return (self._exchange_params_by_currency_id is not None) and (
            self._utility_params_by_good_id is not None
        )

    @property
    def exchange_params_by_currency_id(self) -> ExchangeParams:
        """Get exchange parameter for each currency."""
        if self._exchange_params_by_currency_id is None:
            raise ValueError("ExchangeParams not set!")
        return self._exchange_params_by_currency_id

    @property
    def utility_params_by_good_id(self) -> UtilityParams:
        """Get utility parameter for each good."""
        if self._utility_params_by_good_id is None:
            raise ValueError("UtilityParams not set!")
        return self._utility_params_by_good_id

    def logarithmic_utility(self, quantities_by_good_id: GoodHoldings) -> float:
        """
        Compute agent's utility given her utility function params and a good bundle.

        :param quantities_by_good_id: the good holdings (dictionary) with the identifier (key) and quantity (value) for each good
        :return: utility value
        """
        enforce(self.is_initialized, "Preferences params not set!")
        result = logarithmic_utility(
            self.utility_params_by_good_id, quantities_by_good_id, self._quantity_shift
        )
        return result

    def linear_utility(self, amount_by_currency_id: CurrencyHoldings) -> float:
        """
        Compute agent's utility given her utility function params and a currency bundle.

        :param amount_by_currency_id: the currency holdings (dictionary) with the identifier (key) and quantity (value) for each currency
        :return: utility value
        """
        enforce(self.is_initialized, "Preferences params not set!")
        result = linear_utility(
            self.exchange_params_by_currency_id, amount_by_currency_id
        )
        return result

    def utility(
        self,
        quantities_by_good_id: GoodHoldings,
        amount_by_currency_id: CurrencyHoldings,
    ) -> float:
        """
        Compute the utility given the good and currency holdings.

        :param quantities_by_good_id: the good holdings
        :param amount_by_currency_id: the currency holdings
        :return: the utility value.
        """
        enforce(self.is_initialized, "Preferences params not set!")
        goods_score = self.logarithmic_utility(quantities_by_good_id)
        currency_score = self.linear_utility(amount_by_currency_id)
        score = goods_score + currency_score
        return score

    def marginal_utility(  # pylint: disable=arguments-differ
        self,
        ownership_state: BaseOwnershipState,
        delta_quantities_by_good_id: Optional[GoodHoldings] = None,
        delta_amount_by_currency_id: Optional[CurrencyHoldings] = None,
        **kwargs,
    ) -> float:
        """
        Compute the marginal utility.

        :param ownership_state: the ownership state against which to compute the marginal utility.
        :param delta_quantities_by_good_id: the change in good holdings
        :param delta_amount_by_currency_id: the change in money holdings
        :return: the marginal utility score
        """
        enforce(self.is_initialized, "Preferences params not set!")
        ownership_state = cast(OwnershipState, ownership_state)
        current_goods_score = self.logarithmic_utility(
            ownership_state.quantities_by_good_id
        )
        current_currency_score = self.linear_utility(
            ownership_state.amount_by_currency_id
        )
        new_goods_score = current_goods_score
        new_currency_score = current_currency_score
        if delta_quantities_by_good_id is not None:
            new_quantities_by_good_id = {
                good_id: quantity + delta_quantities_by_good_id[good_id]
                for good_id, quantity in ownership_state.quantities_by_good_id.items()
            }
            new_goods_score = self.logarithmic_utility(new_quantities_by_good_id)
        if delta_amount_by_currency_id is not None:
            new_amount_by_currency_id = {
                currency: amount + delta_amount_by_currency_id[currency]
                for currency, amount in ownership_state.amount_by_currency_id.items()
            }
            new_currency_score = self.linear_utility(new_amount_by_currency_id)
        marginal_utility = (
            new_goods_score
            + new_currency_score
            - current_goods_score
            - current_currency_score
        )
        return marginal_utility

    def utility_diff_from_transaction(
        self, ownership_state: BaseOwnershipState, terms: Terms
    ) -> float:
        """
        Simulate a transaction and get the resulting utility difference (taking into account the fee).

        :param ownership_state: the ownership state against which to apply the transaction.
        :param terms: the transaction terms.
        :return: the score.
        """
        enforce(self.is_initialized, "Preferences params not set!")
        ownership_state = cast(OwnershipState, ownership_state)
        current_score = self.utility(
            quantities_by_good_id=ownership_state.quantities_by_good_id,
            amount_by_currency_id=ownership_state.amount_by_currency_id,
        )
        new_ownership_state = ownership_state.apply_transactions([terms])
        new_score = self.utility(
            quantities_by_good_id=new_ownership_state.quantities_by_good_id,
            amount_by_currency_id=new_ownership_state.amount_by_currency_id,
        )
        score_difference = new_score - current_score
        return score_difference

    def is_utility_enhancing(
        self, ownership_state: BaseOwnershipState, terms: Terms
    ) -> bool:
        """
        Check if the tx is utility enhancing.

        :param ownership_state: the ownership state against which to apply the transaction.
        :param terms: the transaction terms
        :return: whether the transaction is utility enhancing or not
        """
        if self.is_initialized and ownership_state.is_initialized:
            is_utility_enhancing = (
                self.utility_diff_from_transaction(ownership_state, terms) >= 0.0
            )
        else:
            _default_logger.debug(
                "Cannot verify whether transaction improves utility as preferences are not initialized. Assuming it does!"
            )
            is_utility_enhancing = True
        return is_utility_enhancing

    def __copy__(self) -> "Preferences":
        """Copy the object."""
        preferences = Preferences()
        if self.is_initialized:
            preferences._exchange_params_by_currency_id = (
                self.exchange_params_by_currency_id
            )
            preferences._utility_params_by_good_id = self.utility_params_by_good_id
        return preferences


class DecisionMakerHandler(BaseDecisionMakerHandler):
    """This class implements the decision maker."""

    # pylint: disable=import-outside-toplevel
    from packages.fetchai.protocols.signing.dialogues import (  # noqa: F811
        SigningDialogue,
    )
    from packages.fetchai.protocols.signing.dialogues import (  # noqa: F811
        SigningDialogues as BaseSigningDialogues,
    )
    from packages.fetchai.protocols.signing.message import SigningMessage  # noqa: F811
    from packages.fetchai.protocols.state_update.dialogues import (  # noqa: F811
        StateUpdateDialogues as BaseStateUpdateDialogues,
    )
    from packages.fetchai.protocols.state_update.message import (  # noqa: F811
        StateUpdateMessage,
    )

    class SigningDialogues(BaseSigningDialogues):
        """This class keeps track of all oef_search dialogues."""

        def __init__(self, self_address: Address, **kwargs) -> None:
            """
            Initialize dialogues.

            :param self_address: the address of the entity for whom dialogues are maintained
            :return: None
            """

            def role_from_first_message(  # pylint: disable=unused-argument
                message: Message, receiver_address: Address
            ) -> BaseDialogue.Role:
                """Infer the role of the agent from an incoming/outgoing first message

                :param message: an incoming/outgoing first message
                :param receiver_address: the address of the receiving agent
                :return: The role of the agent
                """
                from packages.fetchai.protocols.signing.dialogues import (  # pylint: disable=import-outside-toplevel
                    SigningDialogue,
                )

                return SigningDialogue.Role.DECISION_MAKER

            # pylint: disable=import-outside-toplevel
            from packages.fetchai.protocols.signing.dialogues import (
                SigningDialogues as BaseSigningDialogues,
            )

            BaseSigningDialogues.__init__(
                self,
                self_address=self_address,
                role_from_first_message=role_from_first_message,
                **kwargs,
            )

    class StateUpdateDialogues(BaseStateUpdateDialogues):
        """This class keeps track of all oef_search dialogues."""

        def __init__(self, self_address: Address, **kwargs) -> None:
            """
            Initialize dialogues.

            :param self_address: the address of the entity for whom dialogues are maintained
            :return: None
            """

            def role_from_first_message(  # pylint: disable=unused-argument
                message: Message, receiver_address: Address
            ) -> BaseDialogue.Role:
                """Infer the role of the agent from an incoming/outgoing first message

                :param message: an incoming/outgoing first message
                :param receiver_address: the address of the receiving agent
                :return: The role of the agent
                """
                from packages.fetchai.protocols.state_update.dialogues import (  # noqa: F811 # pylint: disable=import-outside-toplevel
                    StateUpdateDialogue,
                )

                return StateUpdateDialogue.Role.DECISION_MAKER

            # pylint: disable=import-outside-toplevel
            from packages.fetchai.protocols.state_update.dialogues import (  # noqa: F401
                StateUpdateDialogues as BaseStateUpdateDialogues,
            )

            BaseStateUpdateDialogues.__init__(
                self,
                self_address=self_address,
                role_from_first_message=role_from_first_message,
                **kwargs,
            )

    def __init__(self, identity: Identity, wallet: Wallet):
        """
        Initialize the decision maker.

        :param identity: the identity
        :param wallet: the wallet
        """
        kwargs = {
            "goal_pursuit_readiness": GoalPursuitReadiness(),
            "ownership_state": OwnershipState(),
            "preferences": Preferences(),
        }
        super().__init__(
            identity=identity, wallet=wallet, **kwargs,
        )
        self.signing_dialogues = DecisionMakerHandler.SigningDialogues(
            self.self_address
        )
        self.state_update_dialogues = DecisionMakerHandler.StateUpdateDialogues(
            self.self_address
        )

    def handle(self, message: Message) -> None:
        """
        Handle an internal message from the skills.

        :param message: the internal message
        :return: None
        """
        from packages.fetchai.protocols.signing.message import SigningMessage
        from packages.fetchai.protocols.state_update.message import StateUpdateMessage

        if isinstance(message, SigningMessage):
            self._handle_signing_message(message)
        elif isinstance(message, StateUpdateMessage):
            self._handle_state_update_message(message)
        else:  # pragma: no cover
            self.logger.error(
                "[{}]: cannot handle message={} of type={}".format(
                    self.agent_name, message, type(message)
                )
            )

    def _handle_signing_message(self, signing_msg: SigningMessage) -> None:
        """
        Handle a signing message.

        :param signing_msg: the transaction message
        :return: None
        """
        if not self.context.goal_pursuit_readiness.is_ready:
            self.logger.debug(
                "[{}]: Preferences and ownership state not initialized!".format(
                    self.agent_name
                )
            )

        from packages.fetchai.protocols.signing.dialogues import SigningDialogue

        signing_dialogue = cast(
            Optional[SigningDialogue], self.signing_dialogues.update(signing_msg)
        )
        if signing_dialogue is None:  # pragma: no cover
            self.logger.error(
                "[{}]: Could not construct signing dialogue. Aborting!".format(
                    self.agent_name
                )
            )
            return

        from packages.fetchai.protocols.signing.message import SigningMessage

        # check if the transaction is acceptable and process it accordingly
        if signing_msg.performative == SigningMessage.Performative.SIGN_MESSAGE:
            self._handle_message_signing(signing_msg, signing_dialogue)
        elif signing_msg.performative == SigningMessage.Performative.SIGN_TRANSACTION:
            self._handle_transaction_signing(signing_msg, signing_dialogue)
        else:  # pragma: no cover
            self.logger.error(
                "[{}]: Unexpected transaction message performative".format(
                    self.agent_name
                )
            )

    def _handle_message_signing(
        self, signing_msg: SigningMessage, signing_dialogue: SigningDialogue
    ) -> None:
        """
        Handle a message for signing.

        :param signing_msg: the signing message
        :param signing_dialogue: the signing dialogue
        :return: None
        """
        from packages.fetchai.protocols.signing.message import SigningMessage

        performative = SigningMessage.Performative.ERROR
        kwargs = {
            "error_code": SigningMessage.ErrorCode.UNSUCCESSFUL_MESSAGE_SIGNING,
        }  # type: Dict[str, Any]
        if self._is_acceptable_for_signing(signing_msg):
            signed_message = self.wallet.sign_message(
                signing_msg.raw_message.ledger_id,
                signing_msg.raw_message.body,
                signing_msg.raw_message.is_deprecated_mode,
            )
            if signed_message is not None:
                performative = SigningMessage.Performative.SIGNED_MESSAGE
                kwargs.pop("error_code")
                kwargs["signed_message"] = SignedMessage(
                    signing_msg.raw_message.ledger_id,
                    signed_message,
                    signing_msg.raw_message.is_deprecated_mode,
                )
        signing_msg_response = signing_dialogue.reply(
            performative=performative, target_message=signing_msg, **kwargs,
        )
        self.message_out_queue.put(signing_msg_response)

    def _handle_transaction_signing(
        self, signing_msg: SigningMessage, signing_dialogue: SigningDialogue
    ) -> None:
        """
        Handle a transaction for signing.

        :param signing_msg: the signing message
        :param signing_dialogue: the signing dialogue
        :return: None
        """
        from packages.fetchai.protocols.signing.message import SigningMessage

        performative = SigningMessage.Performative.ERROR
        kwargs = {
            "error_code": SigningMessage.ErrorCode.UNSUCCESSFUL_TRANSACTION_SIGNING,
        }  # type: Dict[str, Any]
        if self._is_acceptable_for_signing(signing_msg):
            signed_tx = self.wallet.sign_transaction(
                signing_msg.raw_transaction.ledger_id, signing_msg.raw_transaction.body
            )
            if signed_tx is not None:
                performative = SigningMessage.Performative.SIGNED_TRANSACTION
                kwargs.pop("error_code")
                kwargs["signed_transaction"] = SignedTransaction(
                    signing_msg.raw_transaction.ledger_id, signed_tx
                )
        signing_msg_response = signing_dialogue.reply(
            performative=performative, target_message=signing_msg, **kwargs,
        )
        self.message_out_queue.put(signing_msg_response)

    def _is_acceptable_for_signing(self, signing_msg: SigningMessage) -> bool:
        """
        Check if the tx message is acceptable for signing.

        :param signing_msg: the transaction message
        :return: whether the transaction is acceptable or not
        """
        result = self.context.preferences.is_utility_enhancing(
            self.context.ownership_state, signing_msg.terms
        ) and self.context.ownership_state.is_affordable(signing_msg.terms)
        return result

    def _handle_state_update_message(
        self, state_update_msg: StateUpdateMessage
    ) -> None:
        """
        Handle a state update message.

        :param state_update_message: the state update message
        :return: None
        """
        from packages.fetchai.protocols.state_update.dialogues import (  # noqa: F811
            StateUpdateDialogue,
        )

        state_update_dialogue = cast(
            Optional[StateUpdateDialogue],
            self.state_update_dialogues.update(state_update_msg),
        )
        if state_update_dialogue is None:  # pragma: no cover
            self.logger.error(
                "[{}]: Could not construct state_update dialogue. Aborting!".format(
                    self.agent_name
                )
            )
            return

        from packages.fetchai.protocols.state_update.message import StateUpdateMessage

        if state_update_msg.performative == StateUpdateMessage.Performative.INITIALIZE:
            self.logger.info(
                "[{}]: Applying ownership_state and preferences initialization!".format(
                    self.agent_name
                )
            )
            self.context.ownership_state.set(
                amount_by_currency_id=state_update_msg.amount_by_currency_id,
                quantities_by_good_id=state_update_msg.quantities_by_good_id,
            )
            self.context.preferences.set(
                exchange_params_by_currency_id=state_update_msg.exchange_params_by_currency_id,
                utility_params_by_good_id=state_update_msg.utility_params_by_good_id,
            )
            self.context.goal_pursuit_readiness.update(
                GoalPursuitReadiness.Status.READY
            )
        elif state_update_msg.performative == StateUpdateMessage.Performative.APPLY:
            self.logger.info("[{}]: Applying state update!".format(self.agent_name))
            self.context.ownership_state.apply_delta(
                delta_amount_by_currency_id=state_update_msg.amount_by_currency_id,
                delta_quantities_by_good_id=state_update_msg.quantities_by_good_id,
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Sym link implementation for Linux, MacOS, and Windows."""

import contextlib
import os
from functools import reduce
from pathlib import Path


def make_symlink(link_name: str, target: str):
    """
    Make a symbolic link, cross platform.

    :param link_name: the link name.
    :param target: the target.
    """
    try:
        Path(link_name).unlink()
    except FileNotFoundError:
        pass
    Path(link_name).symlink_to(target, target_is_directory=True)


@contextlib.contextmanager
def cd(path):
    """Change directory with context manager."""
    old_cwd = os.getcwd()
    try:
        os.chdir(path)
        yield
        os.chdir(old_cwd)
    except Exception as e:  # pylint: disable=broad-except  # pragma: nocover
        os.chdir(old_cwd)
        raise e from e


def create_symlink(link_path: Path, target_path: Path, root_path: Path):
    """
    Change directory and call the cross-platform script.

    The working directory must be the parent of the symbolic link name
    when executing 'create_symlink_crossplatform.sh'. Hence, we
    need to translate target_path into the relatve path from the
    symbolic link directory to the target directory.

    So:
    1) from link_path, extract the number of jumps to the parent directory
      in order to reach the repository root directory, and chain many "../" paths.
    2) from target_path, compute the relative path to the root
    3) relative_target_path is just the concatenation of the results from step (1) and (2).


    For instance, given
    - link_path: './directory_1//symbolic_link
    - target_path: './directory_2/target_path

    we want to compute:
    - link_path: 'symbolic_link' (just the last bit)
    - relative_target_path: '../../directory_1/target_path'

    The resulting command on UNIX systems will be:

        cd directory_1 && ln -s ../../directory_1/target_path symbolic_link

    """
    working_directory = link_path.parent
    target_relative_to_root = target_path.relative_to(root_path)
    cwd_relative_to_root = working_directory.relative_to(root_path)
    nb_parents = len(cwd_relative_to_root.parents)
    root_relative_to_cwd = reduce(
        lambda x, y: x / y, [Path("../")] * nb_parents, Path(".")
    )
    link_name = link_path.name
    target = root_relative_to_cwd / target_relative_to_root
    with cd(working_directory.absolute()):
        make_symlink(str(link_name), str(target))
    return 0



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the models of the tac control skill."""

import datetime
import logging
import pprint
from pathlib import Path
from unittest.mock import Mock, patch

import pytest

from aea.crypto.ledger_apis import LedgerApis
from aea.exceptions import AEAEnforceError
from aea.helpers.preference_representations.base import (
    linear_utility,
    logarithmic_utility,
)
from aea.helpers.search.models import Description, Location
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.tac.message import TacMessage
from packages.fetchai.skills.tac_control.game import (
    AGENT_LOCATION_MODEL,
    AGENT_REMOVE_SERVICE_MODEL,
    AGENT_SET_SERVICE_MODEL,
    AgentState,
    Configuration,
    Game,
    Initialization,
    Phase,
    Registration,
    Transaction,
    Transactions,
)
from packages.fetchai.skills.tac_control.parameters import Parameters

from tests.conftest import ROOT_DIR


class TestConfiguration:
    """Test Configuration class of tac control."""

    @classmethod
    def setup(cls):
        """Setup the test class."""
        cls.version_id = "some_version_id"
        cls.tx_fee = 1
        cls.agent_addr_to_name = {
            "agent_address_1": "agent_name_1",
            "agent_address_2": "agent_name_2",
        }
        cls.currency_id_to_name = {"1": "currency_1"}
        cls.good_id_to_name = {"3": "good_1", "4": "good_2"}

        cls.configuration = Configuration(
            cls.version_id,
            cls.tx_fee,
            cls.agent_addr_to_name,
            cls.currency_id_to_name,
            cls.good_id_to_name,
        )

    def test_simple_properties(self):
        """Test the properties of Game class."""
        assert self.configuration.version_id == self.version_id

        assert self.configuration.fee_by_currency_id == {"1": 1}

        assert self.configuration.agent_addr_to_name == self.agent_addr_to_name

        assert self.configuration.currency_id_to_name == self.currency_id_to_name

        assert self.configuration.good_id_to_name == self.good_id_to_name

        assert self.configuration.has_contract_address is False

        with pytest.raises(AEAEnforceError, match="Contract_address not set yet!"):
            assert self.configuration.contract_address

        self.configuration.contract_address = "some_contract_address"

        assert self.configuration.has_contract_address is True
        assert self.configuration.contract_address == "some_contract_address"

        with pytest.raises(AEAEnforceError, match="Contract_address already set!"):
            self.configuration.contract_address = "some_other_contract_address"

    def test_check_consistency_succeeds(self):
        """Test the _check_consistency of Configuration class which succeeds."""
        self.configuration._check_consistency()

    def test_check_consistency_fails_i(self):
        """Test the _check_consistency of Configuration class which fails on version being None."""
        self.configuration._version_id = None
        with pytest.raises(AEAEnforceError, match="A version id must be set."):
            assert self.configuration._check_consistency()

    def test_check_consistency_fails_ii(self):
        """Test the _check_consistency of Configuration class which fails because tx_fee < 0."""
        self.configuration._tx_fee = -5
        with pytest.raises(AEAEnforceError, match="Tx fee must be non-negative."):
            assert self.configuration._check_consistency()

    def test_check_consistency_fails_iii(self):
        """Test the _check_consistency of Configuration class which fails because number of agents is less than 2."""
        self.configuration._agent_addr_to_name = {"agent_address_1": "agent_name_1"}
        with pytest.raises(AEAEnforceError, match="Must have at least two agents."):
            assert self.configuration._check_consistency()

    def test_check_consistency_fails_iv(self):
        """Test the _check_consistency of Configuration class which fails because number of goods is less than 2."""
        self.configuration._good_id_to_name = {"3": "good_1"}
        with pytest.raises(AEAEnforceError, match="Must have at least two goods."):
            assert self.configuration._check_consistency()

    def test_check_consistency_fails_v(self):
        """Test the _check_consistency of Configuration class which fails because number of currencies is not 1."""
        self.configuration._currency_id_to_name = {"1": "currency_1", "2": "currency_2"}
        with pytest.raises(AEAEnforceError, match="Must have exactly one currency."):
            assert self.configuration._check_consistency()

    def test_check_consistency_fails_vi(self):
        """Test the _check_consistency of Configuration class which fails because same id for good and currency."""
        self.configuration._good_id_to_name = {"1": "good_1", "2": "good_2"}
        self.configuration._currency_id_to_name = {"1": "currency_1"}
        with pytest.raises(
            AEAEnforceError, match="Currency id and good ids cannot overlap."
        ):
            assert self.configuration._check_consistency()


class TestInitialization:
    """Test Initialization class of tac control."""

    @classmethod
    def setup(cls):
        """Setup the test class."""
        cls.agent_addr_to_currency_endowments = {
            "agent_address_1": {"1": 10},
            "agent_address_2": {"1": 5},
        }
        cls.agent_addr_to_exchange_params = {
            "agent_address_1": {"1": 1.0},
            "agent_address_2": {"1": 1.5},
        }
        cls.agent_addr_to_good_endowments = {
            "agent_address_1": {"2": 5, "3": 7},
            "agent_address_2": {"2": 4, "3": 6},
        }
        cls.agent_addr_to_utility_params = {
            "agent_address_1": {"2": 1.0, "3": 1.1},
            "agent_address_2": {"2": 1.3, "3": 1.5},
        }
        cls.good_id_to_eq_prices = {"2": 1.7, "3": 1.3}
        cls.agent_addr_to_eq_good_holdings = {
            "agent_address_1": {"2": 1.2, "3": 1.1},
            "agent_address_2": {"2": 1.1, "3": 1.4},
        }
        cls.agent_addr_to_eq_currency_holdings = {
            "agent_address_1": {"1": 1.1},
            "agent_address_2": {"1": 1.2},
        }

        cls.initialization = Initialization(
            cls.agent_addr_to_currency_endowments,
            cls.agent_addr_to_exchange_params,
            cls.agent_addr_to_good_endowments,
            cls.agent_addr_to_utility_params,
            cls.good_id_to_eq_prices,
            cls.agent_addr_to_eq_good_holdings,
            cls.agent_addr_to_eq_currency_holdings,
        )

    def test_simple_properties(self):
        """Test the properties of Game class."""
        assert (
            self.initialization.agent_addr_to_currency_endowments
            == self.agent_addr_to_currency_endowments
        )
        assert (
            self.initialization.agent_addr_to_exchange_params
            == self.agent_addr_to_exchange_params
        )
        assert (
            self.initialization.agent_addr_to_good_endowments
            == self.agent_addr_to_good_endowments
        )
        assert (
            self.initialization.agent_addr_to_utility_params
            == self.agent_addr_to_utility_params
        )
        assert self.initialization.good_id_to_eq_prices == self.good_id_to_eq_prices
        assert (
            self.initialization.agent_addr_to_eq_good_holdings
            == self.agent_addr_to_eq_good_holdings
        )
        assert (
            self.initialization.agent_addr_to_eq_currency_holdings
            == self.agent_addr_to_eq_currency_holdings
        )

    def test_check_consistency_succeeds(self):
        """Test the _check_consistency of Configuration class which succeeds."""
        self.initialization._check_consistency()

    def test_check_consistency_fails_i(self):
        """Test the _check_consistency of Configuration class which fails because currency endowments are negative."""
        self.initialization._agent_addr_to_currency_endowments = {
            "agent_address_1": {"1": -1},
            "agent_address_2": {"1": 5},
        }
        with pytest.raises(
            AEAEnforceError, match="Currency endowments must be non-negative."
        ):
            assert self.initialization._check_consistency()

    def test_check_consistency_fails_ii(self):
        """Test the _check_consistency of Configuration class which fails because ExchangeParams are not strictly positive."""
        self.initialization._agent_addr_to_exchange_params = {
            "agent_address_1": {"1": 0.0},
            "agent_address_2": {"1": -1.2},
        }
        with pytest.raises(
            AEAEnforceError, match="ExchangeParams must be strictly positive."
        ):
            assert self.initialization._check_consistency()

    def test_check_consistency_fails_iii(self):
        """Test the _check_consistency of Configuration class which fails because Good endowments are not strictly positive."""
        self.initialization._agent_addr_to_good_endowments = {
            "agent_address_1": {"2": 0, "3": -1},
            "agent_address_2": {"2": -7, "3": 0},
        }
        with pytest.raises(
            AEAEnforceError, match="Good endowments must be strictly positive."
        ):
            assert self.initialization._check_consistency()

    def test_check_consistency_fails_iv(self):
        """Test the _check_consistency of Configuration class which fails because UtilityParams are not strictly positive."""
        self.initialization._agent_addr_to_utility_params = {
            "agent_address_1": {"2": 0, "3": -7},
            "agent_address_2": {"2": -4, "3": 0},
        }
        with pytest.raises(
            AEAEnforceError, match="UtilityParams must be strictly positive."
        ):
            assert self.initialization._check_consistency()

    def test_check_consistency_fails_v(self):
        """Test the _check_consistency of Configuration class which fails because lengths of endowments are not the same."""
        self.initialization._agent_addr_to_currency_endowments = {
            "agent_address_1": {"1": 10},
        }
        with pytest.raises(
            AEAEnforceError, match="Length of endowments must be the same."
        ):
            assert self.initialization._check_consistency()

    def test_check_consistency_fails_vi(self):
        """Test the _check_consistency of Configuration class which fails because lengths of params are not the same."""
        self.initialization._agent_addr_to_exchange_params = {
            "agent_address_1": {"1": 1.0},
        }
        with pytest.raises(AEAEnforceError, match="Length of params must be the same."):
            assert self.initialization._check_consistency()

    def test_check_consistency_fails_vii(self):
        """Test the _check_consistency of Configuration class which fails because length of eq_prices and elements of eq_good_holdings are not the same."""
        self.initialization._good_id_to_eq_prices = {"2": 1.7}
        with pytest.raises(
            AEAEnforceError,
            match="Length of eq_prices and an element of eq_good_holdings must be the same.",
        ):
            assert self.initialization._check_consistency()

    def test_check_consistency_fails_viii(self):
        """Test the _check_consistency of Configuration class which fails because length of eq_good_holdings and eq_currency_holdings are not the same."""
        self.initialization._agent_addr_to_eq_currency_holdings = {
            "agent_address_1": {"1": 1.1},
        }
        with pytest.raises(
            AEAEnforceError,
            match="Length of eq_good_holdings and eq_currency_holdings must be the same.",
        ):
            assert self.initialization._check_consistency()

    def test_check_consistency_fails_ix(self):
        """Test the _check_consistency of Configuration class which fails because exchange_params and currency_endowments have different number of rows."""
        self.initialization._agent_addr_to_currency_endowments = {
            "agent_address_1": {"1": 10, "2": 11},
            "agent_address_2": {"1": 5},
        }
        with pytest.raises(
            AEAEnforceError,
            match="Dimensions for exchange_params and currency_endowments rows must be the same.",
        ):
            assert self.initialization._check_consistency()

    def test_check_consistency_fails_x(self):
        """Test the _check_consistency of Configuration class which fails because utility_params and rows have different number of rows."""
        self.initialization._agent_addr_to_good_endowments = {
            "agent_address_1": {"2": 5, "3": 7, "4": 8},
            "agent_address_2": {"2": 4, "3": 6},
        }
        with pytest.raises(
            AEAEnforceError,
            match="Dimensions for utility_params and good_endowments rows must be the same.",
        ):
            assert self.initialization._check_consistency()


class TestTransaction:
    """Test Initialization class of tac control."""

    @classmethod
    def setup(cls):
        """Setup the test class."""
        cls.ledger_id = "ethereum"
        cls.sender_address = "some_sender_address"
        cls.counterparty_address = "some_counterparty_address"
        cls.amount_by_currency_id = {"1": 10}
        cls.quantities_by_good_id = {"2": 5, "5": 10}
        cls.is_sender_payable_tx_fee = True
        cls.nonce = "some_nonce"
        cls.fee_by_currency_id = {"1": 1}
        cls.sender_signature = "some_sender_signature"
        cls.counterparty_signature = "some_counterparty_signature"

        cls.transaction = Transaction(
            cls.ledger_id,
            cls.sender_address,
            cls.counterparty_address,
            cls.amount_by_currency_id,
            cls.quantities_by_good_id,
            cls.is_sender_payable_tx_fee,
            cls.nonce,
            cls.fee_by_currency_id,
            cls.sender_signature,
            cls.counterparty_signature,
        )

    def test_simple_properties(self):
        """Test the properties of Game class."""
        assert self.transaction.sender_signature == self.sender_signature
        assert self.transaction.counterparty_signature == self.counterparty_signature

    def test_has_matching_signatures_succeeds(self):
        """Test the has_matching_signatures method of Transaction class where the two addresses appear in the hash."""
        with patch.object(
            LedgerApis,
            "recover_message",
            return_value=(self.sender_address, self.counterparty_address),
        ):
            assert self.transaction.has_matching_signatures() is True

    def test_has_matching_signatures_fails_sender_not_in_hash(self):
        """Test the has_matching_signatures method of Transaction class where the sender address does not appear in the hash."""
        with patch.object(
            LedgerApis, "recover_message", return_value=(self.counterparty_address,)
        ):
            assert self.transaction.has_matching_signatures() is False

    def test_has_matching_signatures_fails_counterparty_not_in_hash(self):
        """Test the has_matching_signatures method of Transaction class where the counterparty addresses does not appear in the hash."""
        with patch.object(
            LedgerApis, "recover_message", return_value=(self.sender_address,)
        ):
            assert self.transaction.has_matching_signatures() is False

    def test_has_matching_signatures_fails_sender_and_counterparty_not_in_hash(self):
        """Test the has_matching_signatures method of Transaction class where the sender and counterparty addresses do not appear in the hash."""
        with patch.object(LedgerApis, "recover_message", return_value=tuple()):
            assert self.transaction.has_matching_signatures() is False

    def test_from_message(self):
        """Test the from_message method of Transaction class."""
        ledger_id = "some_ledger_id"
        sender_address = "some_sender_address"
        counterparty_address = "some_counterparty_address"
        amount_by_currency_id = {"FET": 10}
        fee_by_currency_id = {"FET": 2}
        quantities_by_good_id = {"G1": -1}
        nonce = "some_nonce"
        sender_signature = "some_signature"
        counterparty_signature = "some_other_signature"

        tx_id = Transaction.get_hash(
            ledger_id,
            sender_address=sender_address,
            counterparty_address=counterparty_address,
            good_ids=["G1"],
            sender_supplied_quantities=[0],
            counterparty_supplied_quantities=[1],
            sender_payable_amount=0,
            counterparty_payable_amount=10,
            nonce=nonce,
        )

        tx = Transaction.from_message(
            TacMessage(
                performative=TacMessage.Performative.TRANSACTION,
                transaction_id=tx_id,
                ledger_id=ledger_id,
                sender_address=sender_address,
                counterparty_address=counterparty_address,
                amount_by_currency_id=amount_by_currency_id,
                fee_by_currency_id=fee_by_currency_id,
                quantities_by_good_id=quantities_by_good_id,
                nonce=nonce,
                sender_signature=sender_signature,
                counterparty_signature=counterparty_signature,
            )
        )

        assert tx.ledger_id == ledger_id
        assert tx.sender_address == sender_address
        assert tx.counterparty_address == counterparty_address
        assert tx.amount_by_currency_id == amount_by_currency_id
        assert tx.fee_by_currency_id == fee_by_currency_id
        assert tx.quantities_by_good_id == quantities_by_good_id
        assert tx.nonce == nonce
        assert tx.sender_signature == sender_signature
        assert tx.counterparty_signature == counterparty_signature

    def test__eq__(self):
        """Test the __eq__ method of Transaction class."""
        equal_transaction = Transaction(
            self.ledger_id,
            self.sender_address,
            self.counterparty_address,
            self.amount_by_currency_id,
            self.quantities_by_good_id,
            self.is_sender_payable_tx_fee,
            self.nonce,
            self.fee_by_currency_id,
            self.sender_signature,
            self.counterparty_signature,
        )
        assert self.transaction.__eq__(equal_transaction) is True

        not_equal_transaction = Transaction(
            self.ledger_id,
            "some_different_sender_address",
            self.counterparty_address,
            self.amount_by_currency_id,
            self.quantities_by_good_id,
            self.is_sender_payable_tx_fee,
            self.nonce,
            self.fee_by_currency_id,
            self.sender_signature,
            self.counterparty_signature,
        )
        assert self.transaction.__eq__(not_equal_transaction) is False


class TestAgentState:
    """Test AgentState class of tac control."""

    @classmethod
    def setup(cls):
        """Setup the test class."""
        cls.agent_address = "sender_address"
        cls.amount_by_currency_id = {"1": 10}
        cls.quantities_by_good_id = {"2": 1, "3": 2}
        cls.exchange_params_by_currency_id = {"1": 1.0}
        cls.utility_params_by_good_id = {"2": 1.0, "3": 1.5}

        cls.agent_state = AgentState(
            cls.agent_address,
            cls.amount_by_currency_id,
            cls.exchange_params_by_currency_id,
            cls.quantities_by_good_id,
            cls.utility_params_by_good_id,
        )

        cls.ledger_id = "ethereum"
        cls.sender_address = cls.agent_address
        cls.counterparty_address = "some_counterparty_address"
        cls.tx_amount_by_currency_id = {"1": 10}
        cls.tx_quantities_by_good_id = {"2": -1, "3": -2}
        cls.is_sender_payable_tx_fee = True
        cls.nonce = "some_nonce"
        cls.fee_by_currency_id = {"1": 1}
        cls.sender_signature = "some_sender_signature"
        cls.counterparty_signature = "some_counterparty_signature"
        cls.transaction_1 = Transaction(
            cls.ledger_id,
            cls.sender_address,
            cls.counterparty_address,
            cls.tx_amount_by_currency_id,
            cls.tx_quantities_by_good_id,
            cls.is_sender_payable_tx_fee,
            cls.nonce,
            cls.fee_by_currency_id,
            cls.sender_signature,
            cls.counterparty_signature,
        )

        cls.amount_by_currency_id_2 = {"1": -9}
        cls.quantities_by_good_id_2 = {"2": 1, "3": 2}
        cls.transaction_2 = Transaction(
            cls.ledger_id,
            cls.sender_address,
            cls.counterparty_address,
            cls.amount_by_currency_id_2,
            cls.quantities_by_good_id_2,
            cls.is_sender_payable_tx_fee,
            cls.nonce,
            cls.fee_by_currency_id,
            cls.sender_signature,
            cls.counterparty_signature,
        )

    def test_simple_properties(self):
        """Test the properties of AgentState class."""
        assert self.agent_state.agent_address == self.agent_address
        assert self.agent_state.amount_by_currency_id == self.amount_by_currency_id
        assert (
            self.agent_state.exchange_params_by_currency_id
            == self.exchange_params_by_currency_id
        )
        assert self.agent_state.quantities_by_good_id == self.quantities_by_good_id
        assert (
            self.agent_state.utility_params_by_good_id == self.utility_params_by_good_id
        )

    def test_get_score(self):
        """Test the get_score of AgentState class."""
        assert self.agent_state.get_score() == logarithmic_utility(
            self.utility_params_by_good_id, self.quantities_by_good_id
        ) + linear_utility(
            self.exchange_params_by_currency_id, self.amount_by_currency_id
        )

    def test_is_consistent_transaction_succeeds(self):
        """Test the is_consistent_transaction of AgentState class where it returns True."""
        assert self.agent_state.is_consistent_transaction(self.transaction_1) is True

    def test_is_consistent_transaction_fails_i(self):
        """Test the is_consistent_transaction of AgentState class where it fails because agent address is not sender/counterparty."""
        self.transaction_1._sender_address = "some_sender_address"
        assert self.agent_state.is_consistent_transaction(self.transaction_1) is False

    def test_is_consistent_transaction_fails_ii(self):
        """Test the is_consistent_transaction of AgentState class where it fails because tx is not single currency."""
        self.transaction_1._amount_by_currency_id = {"1": 10, "2": 20}
        assert self.agent_state.is_consistent_transaction(self.transaction_1) is False

    def test_is_consistent_transaction_fails_iii(self):
        """Test the is_consistent_transaction of AgentState class where it fails because there is no exchange of wealth."""
        self.transaction_1._amount_by_currency_id = {"1": 0}
        self.transaction_1._quantities_by_good_id = {"2": 0, "3": 0}
        assert self.agent_state.is_consistent_transaction(self.transaction_1) is False

    def test_is_consistent_transaction_fails_iv(self):
        """Test the is_consistent_transaction of AgentState class where it fails because sender does not have enough funds."""
        self.transaction_1._amount_by_currency_id = {"1": -11}
        self.transaction_1._quantities_by_good_id = {"2": 1, "3": 0}
        assert self.agent_state.is_consistent_transaction(self.transaction_1) is False

    def test_is_consistent_transaction_succeeds_iv(self):
        """Test the is_consistent_transaction of AgentState class where it succeeds and sender does have enough funds."""
        self.transaction_1._amount_by_currency_id = {"1": -9}
        self.transaction_1._quantities_by_good_id = {"2": 1, "3": 0}
        assert self.agent_state.is_consistent_transaction(self.transaction_1) is True

    def test_is_consistent_transaction_fails_v(self):
        """Test the is_consistent_transaction of AgentState class where it fails because counterparty does not have enough goods."""
        self.transaction_1._counterparty_address = self.agent_address
        self.transaction_1._sender_address = "some_sender_address"

        self.transaction_1._amount_by_currency_id = {"1": -10}
        self.transaction_1._quantities_by_good_id = {"2": 2, "3": 2}
        assert self.agent_state.is_consistent_transaction(self.transaction_1) is False

    def test_is_consistent_transaction_succeeds_v(self):
        """Test the is_consistent_transaction of AgentState class where it succeeds and counterparty does have enough goods."""
        self.transaction_1._counterparty_address = self.agent_address
        self.transaction_1._sender_address = "some_sender_address"

        self.transaction_1._amount_by_currency_id = {"1": -10}
        self.transaction_1._quantities_by_good_id = {"2": 1, "3": 2}
        assert self.agent_state.is_consistent_transaction(self.transaction_1) is True

    def test_is_consistent_transaction_fails_vi(self):
        """Test the is_consistent_transaction of AgentState class where it fails because sender does not have enough goods."""
        self.transaction_1._amount_by_currency_id = {"1": 10}
        self.transaction_1._quantities_by_good_id = {"2": -2, "3": -2}
        assert self.agent_state.is_consistent_transaction(self.transaction_1) is False

    def test_is_consistent_transaction_succeeds_vi(self):
        """Test the is_consistent_transaction of AgentState class where it succeeds and sender does have enough goods."""
        self.transaction_1._amount_by_currency_id = {"1": 10}
        self.transaction_1._quantities_by_good_id = {"2": -1, "3": -2}
        assert self.agent_state.is_consistent_transaction(self.transaction_1) is True

    def test_is_consistent_transaction_fails_vii(self):
        """Test the is_consistent_transaction of AgentState class where it fails because counterparty does not have enough funds."""
        self.transaction_1._counterparty_address = self.agent_address
        self.transaction_1._sender_address = "some_sender_address"

        self.transaction_1._amount_by_currency_id = {"1": 11}
        self.transaction_1._quantities_by_good_id = {"2": -1, "3": -2}
        assert self.agent_state.is_consistent_transaction(self.transaction_1) is False

    def test_is_consistent_transaction_succeeds_vii(self):
        """Test the is_consistent_transaction of AgentState class where it succeeds and counterparty does have enough funds."""
        self.transaction_1._counterparty_address = self.agent_address
        self.transaction_1._sender_address = "some_sender_address"

        self.transaction_1._amount_by_currency_id = {"1": 9}
        self.transaction_1._quantities_by_good_id = {"2": -1, "3": -2}
        assert self.agent_state.is_consistent_transaction(self.transaction_1) is True

    def test_is_consistent_transaction_fails_viii(self):
        """Test the is_consistent_transaction of AgentState class where it fails because inconsistent values."""
        self.transaction_1._amount_by_currency_id = {"1": -11}
        self.transaction_1._quantities_by_good_id = {"2": -1, "3": -2}
        assert self.agent_state.is_consistent_transaction(self.transaction_1) is False

    def test_apply(self):
        """Test the apply of AgentState class."""
        new_agent_state = self.agent_state.apply(
            [self.transaction_1, self.transaction_2]
        )
        assert new_agent_state.amount_by_currency_id == {"1": 11}
        assert new_agent_state.quantities_by_good_id == {"2": 1, "3": 2}

    def test_update_sender_i(self):
        """Test the update of AgentState class where agent is tx sender."""
        self.agent_state.update(self.transaction_1)
        assert self.agent_state.amount_by_currency_id == {"1": 20}
        assert self.agent_state.quantities_by_good_id == {"2": 0, "3": 0}

    def test_update_sender_ii(self):
        """Test the update of AgentState class where agent is tx sender."""
        self.agent_state.update(self.transaction_2)
        assert self.agent_state.amount_by_currency_id == {"1": 1}
        assert self.agent_state.quantities_by_good_id == {"2": 2, "3": 4}

    def test_update_counterparty_i(self):
        """Test the update of AgentState class where agent is tx counterparty."""
        # setup
        self.transaction_1._sender_address = "some_sender_address"
        self.transaction_1._counterparty_address = self.agent_address

        # operation
        self.agent_state.update(self.transaction_1)

        # after
        assert self.agent_state.amount_by_currency_id == {"1": 0}
        assert self.agent_state.quantities_by_good_id == {"2": 2, "3": 4}

    def test_update_counterparty_ii(self):
        """Test the update of AgentState class where agent is tx counterparty."""
        # setup
        self.transaction_2._sender_address = "some_sender_address"
        self.transaction_2._counterparty_address = self.agent_address

        # operation
        self.agent_state.update(self.transaction_2)

        # after
        assert self.agent_state.amount_by_currency_id == {"1": 19}
        assert self.agent_state.quantities_by_good_id == {"2": 0, "3": 0}

    def test__copy__(self):
        """Test the __copy__ of AgentState class."""
        new_agent_state = self.agent_state.__copy__()
        assert new_agent_state == self.agent_state

    def test__str__(self):
        """Test the __str__ of AgentState class."""
        agent_state_str = self.agent_state.__str__()
        assert agent_state_str == "AgentState{}".format(
            pprint.pformat(
                {
                    "agent_address": self.agent_state.agent_address,
                    "amount_by_currency_id": self.agent_state.amount_by_currency_id,
                    "exchange_params_by_currency_id": self.agent_state.exchange_params_by_currency_id,
                    "quantities_by_good_id": self.agent_state.quantities_by_good_id,
                    "utility_params_by_good_id": self.agent_state.utility_params_by_good_id,
                }
            )
        )

    def test__eq__(self):
        """Test the __eq__ of AgentState class."""
        another_agent_state = AgentState(
            self.agent_address,
            self.amount_by_currency_id,
            self.exchange_params_by_currency_id,
            self.quantities_by_good_id,
            self.utility_params_by_good_id,
        )

        assert self.agent_state.__eq__(another_agent_state) is True


class TestTransactions:
    """Test Initialization class of tac control."""

    @classmethod
    def setup(cls):
        """Setup the test class."""
        cls.transactions = Transactions()

    def test_simple_properties(self):
        """Test the properties of Game class."""
        assert self.transactions.confirmed == {}
        assert self.transactions.confirmed_per_agent == {}

    def test_add(self):
        """Test the add of Transactions class which succeeds."""
        ledger_id = "ethereum"
        sender_address = "some_agent_address"
        counterparty_address = "some_counterparty_address"
        tx_amount_by_currency_id = {"1": 10}
        tx_quantities_by_good_id = {"2": -1, "3": -2}
        is_sender_payable_tx_fee = True
        nonce = "some_nonce"
        fee_by_currency_id = {"1": 1}
        sender_signature = "some_sender_signature"
        counterparty_signature = "some_counterparty_signature"
        transaction = Transaction(
            ledger_id,
            sender_address,
            counterparty_address,
            tx_amount_by_currency_id,
            tx_quantities_by_good_id,
            is_sender_payable_tx_fee,
            nonce,
            fee_by_currency_id,
            sender_signature,
            counterparty_signature,
        )

        mocked_now = datetime.datetime.strptime("01 01 2020  00:01", "%d %m %Y %H:%M")

        datetime_mock = Mock(wraps=datetime.datetime)
        datetime_mock.now.return_value = mocked_now

        with patch("datetime.datetime", new=datetime_mock):
            self.transactions.add(transaction)

        assert self.transactions.confirmed[mocked_now] == transaction
        assert (
            self.transactions.confirmed_per_agent[sender_address][mocked_now]
            == transaction
        )
        assert (
            self.transactions.confirmed_per_agent[counterparty_address][mocked_now]
            == transaction
        )


class TestRegistration:
    """Test Registration class of tac control."""

    @classmethod
    def setup(cls):
        """Setup the test class."""
        cls.registration = Registration()

    def test_simple_properties(self):
        """Test the properties of Game class."""
        assert self.registration.agent_addr_to_name == {}
        assert self.registration.nb_agents == 0

    def test_register_agent(self):
        """Test the register_agent of Registration class which succeeds."""
        agent_addr = "some_agent_address"
        agent_name = "some_agent_name"

        self.registration.register_agent(agent_addr, agent_name)

        assert self.registration.agent_addr_to_name == {agent_addr: agent_name}
        assert self.registration.nb_agents == 1

    def test_unregister_agent(self):
        """Test the unregister_agent of Registration class which succeeds."""
        agent_addr = "some_agent_address"
        agent_name = "some_agent_name"

        self.registration.register_agent(agent_addr, agent_name)
        assert self.registration.agent_addr_to_name == {agent_addr: agent_name}
        assert self.registration.nb_agents == 1

        self.registration.unregister_agent(agent_addr)
        assert self.registration.agent_addr_to_name == {}
        assert self.registration.nb_agents == 0


class TestGame(BaseSkillTestCase):
    """Test Game class of tac control."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_control")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.amount_by_currency_id = {"FET": 10}
        cls.exchange_params_by_currency_id = {"FET": 1.0}
        cls.quantities_by_good_id = {"G1": 1, "G2": 2}
        cls.utility_params_by_good_id = {"G1": 1.0, "G2": 1.5}
        cls.game = Game(name="Game", skill_context=cls._skill.skill_context)
        cls._skill.skill_context.parameters = Parameters(
            ledger_id="",
            contract_address=None,
            good_ids=[],
            currency_ids=[],
            min_nb_agents=2,
            money_endowment=200,
            nb_goods=9,
            nb_currencies=1,
            tx_fee=1,
            base_good_endowment=2,
            lower_bound_factor=1,
            upper_bound_factor=1,
            registration_start_time="01 01 2020  00:01",
            registration_timeout=60,
            item_setup_timeout=60,
            competition_timeout=300,
            inactivity_timeout=30,
            whitelist=[],
            location={"longitude": 0.1270, "latitude": 51.5194},
            service_data={"key": "tac", "value": "v1"},
            name="parameters",
            skill_context=cls._skill.skill_context,
        )
        cls.game._conf = "stub"

    def test_simple_properties(self):
        """Test the properties of Game class."""
        self.game._conf = None
        # phase
        assert self.game.phase == Phase.PRE_GAME

        with patch.object(self.game.context.logger, "log") as mock_logger:
            self.game.phase = Phase.GAME
        mock_logger.assert_any_call(logging.DEBUG, f"Game phase set to: {Phase.GAME}")
        assert self.game.phase == Phase.GAME

        # registration
        assert self.game.registration.nb_agents == 0
        assert self.game.registration.agent_addr_to_name == {}

        # conf
        with pytest.raises(
            AEAEnforceError, match="Call create before calling configuration."
        ):
            assert self.game.conf
        conf = Configuration(
            "some_version_id",
            1,
            {"ag_1_add": "ag_1", "ag_2_add": "ag_2"},
            {"FET": "fetch"},
            {"G_1": "good_1", "G_2": "good_2"},
        )
        self.game._conf = conf
        assert self.game.conf == conf

        # initialization
        with pytest.raises(
            AEAEnforceError, match="Call create before calling initialization."
        ):
            assert self.game.initialization
        init = Initialization({}, {}, {}, {}, {}, {}, {})
        self.game._initialization = init
        assert self.game.initialization == init

        # initial_agent_states
        with pytest.raises(
            AEAEnforceError, match="Call create before calling initial_agent_states."
        ):
            assert self.game.initial_agent_states
        ias = {}
        self.game._initial_agent_states = ias
        assert self.game.initial_agent_states == ias

        # current_agent_states
        with pytest.raises(
            AEAEnforceError, match="Call create before calling current_agent_states."
        ):
            assert self.game.current_agent_states
        cas = {}
        self.game._current_agent_states = cas
        assert self.game.current_agent_states == cas

        # transactions
        tx = Transactions()
        self.game._transactions = tx
        assert self.game.transactions == tx

        # is_allowed_to_mint
        assert self.game.is_allowed_to_mint is True
        self.game.is_allowed_to_mint = False
        assert self.game.is_allowed_to_mint is False

    def test_create_succeeds(self):
        """Test the create method of the Game class which succeeds."""
        self.game.phase = Phase.PRE_GAME
        with patch.object(self.game, "_generate") as mock_generate:
            self.game.create()

        assert self.game.phase == Phase.GAME_SETUP
        mock_generate.assert_called_once()

    def test_create_fails(self):
        """Test the create method of the Game class which fails because phase is Game."""
        self.game.phase = Phase.GAME
        with pytest.raises(AEAEnforceError, match="A game phase is already active."):
            with patch.object(self.game, "_generate"):
                self.game.create()

        assert self.game.phase == Phase.GAME

    def test_get_next_agent_state_for_minting(self):
        """Test the get_next_agent_state_for_minting method of the Game class."""
        agent_state = AgentState(
            "some_address_1",
            self.amount_by_currency_id,
            self.exchange_params_by_currency_id,
            self.quantities_by_good_id,
            self.utility_params_by_good_id,
        )
        self.game._initial_agent_states = {"ag1": agent_state}
        self.game._already_minted_agents = []

        actual_agent_state = self.game.get_next_agent_state_for_minting()
        assert actual_agent_state == agent_state

        self.game._already_minted_agents = ["ag1"]
        actual_agent_state = self.game.get_next_agent_state_for_minting()
        assert actual_agent_state is None

    def test_create_generate(self):
        """Test the _generate method of the Game class."""
        # before
        assert self.game._conf == "stub"
        assert self.game._initialization is None
        assert self.game._initial_agent_states is None
        assert self.game._current_agent_states is None

        agent_addr_1 = "some_agent_address_1"
        agent_name_1 = "some_agent_name_1"
        agent_addr_2 = "some_agent_address_2"
        agent_name_2 = "some_agent_name_2"
        self.game.registration.register_agent(agent_addr_1, agent_name_1)
        self.game.registration.register_agent(agent_addr_2, agent_name_2)

        # operation
        self.game._generate()

        # after
        assert self.game._conf != "stub"
        assert self.game.initialization is not None
        assert self.game._initial_agent_states is not None
        assert self.game._current_agent_states is not None

    def test_holdings_summary(self):
        """Test the holdings_summary method of the Game class."""
        # before
        agent_address_1 = "agent_address_1"
        agent_address_2 = "agent_address_2"

        self.game._conf = Configuration(
            "some_version_id",
            1,
            {agent_address_1: "agent_name_1", agent_address_2: "agent_name_2"},
            {"1": "currency_1"},
            {"2": "good_1", "3": "good_2"},
        )
        agent_state_1 = AgentState(
            agent_address_1,
            {"1": 10},
            {"1": 1.0},
            {"2": 1, "3": 2},
            {"2": 1.0, "3": 1.5},
        )
        agent_state_2 = AgentState(
            agent_address_2,
            {"1": 10},
            {"1": 1.0},
            {"2": 1, "3": 2},
            {"2": 1.0, "3": 1.5},
        )

        self.game._current_agent_states = {
            agent_address_1: agent_state_1,
            agent_address_2: agent_state_2,
        }
        expected_holding_summary = (
            "\nCurrent good & money allocation & score: \n"
            "- agent_name_1:\n"
            "    good_1: 1\n"
            "    good_2: 2\n"
            "    currency_1: 10\n"
            "    score: 12.34\n"
            "- agent_name_2:\n"
            "    good_1: 1\n"
            "    good_2: 2\n"
            "    currency_1: 10\n"
            "    score: 12.34\n\n"
        )

        # operation
        holding_summary = self.game.holdings_summary

        # after
        assert holding_summary == expected_holding_summary

    def test_equilibrium_summary(self):
        """Test the equilibrium_summary method of the Game class."""
        # before
        agent_address_1 = "agent_address_1"
        agent_address_2 = "agent_address_2"

        self.game._conf = Configuration(
            "some_version_id",
            1,
            {agent_address_1: "agent_name_1", agent_address_2: "agent_name_2"},
            {"1": "currency_1"},
            {"2": "good_1", "3": "good_2"},
        )

        self.game._initialization = Initialization(
            {agent_address_1: {"1": 10}, agent_address_2: {"1": 5}},
            {agent_address_1: {"1": 1.0}, agent_address_2: {"1": 1.5}},
            {agent_address_1: {"2": 5, "3": 7}, agent_address_2: {"2": 4, "3": 6}},
            {
                agent_address_1: {"2": 1.0, "3": 1.1},
                agent_address_2: {"2": 1.3, "3": 1.5},
            },
            {"2": 1.7, "3": 1.3},
            {
                agent_address_1: {"2": 1.2, "3": 1.1},
                agent_address_2: {"2": 1.1, "3": 1.4},
            },
            {agent_address_1: {"1": 1.1}, agent_address_2: {"1": 1.2}},
        )
        expected_equilibrium_summary = (
            "\nEquilibrium prices: \n"
            "good_1 1.7\n"
            "good_2 1.3\n\n"
            "Equilibrium good allocation: \n"
            "- agent_name_1:\n"
            "    good_1: 1.2\n"
            "    good_2: 1.1\n"
            "- agent_name_2:\n"
            "    good_1: 1.1\n"
            "    good_2: 1.4\n\n"
            "Equilibrium money allocation: \n"
            "agent_name_1 {'1': 1.1}\n"
            "agent_name_2 {'1': 1.2}\n\n"
        )
        # operation
        equilibrium_summary = self.game.equilibrium_summary

        # after
        assert equilibrium_summary == expected_equilibrium_summary

    def test_is_transaction_valid_succeeds(self):
        """Test the is_transaction_valid method of the Game class which succeeds."""
        # before
        agent_address_1 = "agent_address_1"
        agent_address_2 = "agent_address_2"

        tx = Transaction(
            "ethereum",
            agent_address_1,
            agent_address_2,
            {"1": 10},
            {"2": 5, "5": 10},
            True,
            "some_nonce",
            {"1": 1},
            "some_sender_signature",
            "some_counterparty_signature",
        )

        agent_state_1 = AgentState(
            agent_address_1,
            {"1": 10},
            {"1": 1.0},
            {"2": 1, "3": 2},
            {"2": 1.0, "3": 1.5},
        )
        agent_state_2 = AgentState(
            agent_address_2,
            {"1": 10},
            {"1": 1.0},
            {"2": 1, "3": 2},
            {"2": 1.0, "3": 1.5},
        )

        self.game._current_agent_states = {
            agent_address_1: agent_state_1,
            agent_address_2: agent_state_2,
        }

        # operation
        with patch.object(Transaction, "has_matching_signatures", return_value=True):
            with patch.object(
                AgentState, "is_consistent_transaction", return_value=True
            ):
                assert self.game.is_transaction_valid(tx) is True

    def test_is_transaction_valid_fails_not_matching_signatures(self):
        """Test the is_transaction_valid method of the Game class which fails because the signatures do no match."""
        # before
        agent_address_1 = "agent_address_1"
        agent_address_2 = "agent_address_2"

        tx = Transaction(
            "ethereum",
            agent_address_1,
            agent_address_2,
            {"1": 10},
            {"2": 5, "5": 10},
            True,
            "some_nonce",
            {"1": 1},
            "some_sender_signature",
            "some_counterparty_signature",
        )

        agent_state_1 = AgentState(
            agent_address_1,
            {"1": 10},
            {"1": 1.0},
            {"2": 1, "3": 2},
            {"2": 1.0, "3": 1.5},
        )
        agent_state_2 = AgentState(
            agent_address_2,
            {"1": 10},
            {"1": 1.0},
            {"2": 1, "3": 2},
            {"2": 1.0, "3": 1.5},
        )

        self.game._current_agent_states = {
            agent_address_1: agent_state_1,
            agent_address_2: agent_state_2,
        }

        # operation
        with patch.object(Transaction, "has_matching_signatures", return_value=False):
            with patch.object(
                AgentState, "is_consistent_transaction", return_value=True
            ):
                assert self.game.is_transaction_valid(tx) is False

    def test_is_transaction_valid_fails_tx_inconsistent(self):
        """Test the is_transaction_valid method of the Game class which fails because transactions are inconsistent."""
        # before
        agent_address_1 = "agent_address_1"
        agent_address_2 = "agent_address_2"

        tx = Transaction(
            "ethereum",
            agent_address_1,
            agent_address_2,
            {"1": 10},
            {"2": 5, "5": 10},
            True,
            "some_nonce",
            {"1": 1},
            "some_sender_signature",
            "some_counterparty_signature",
        )

        agent_state_1 = AgentState(
            agent_address_1,
            {"1": 10},
            {"1": 1.0},
            {"2": 1, "3": 2},
            {"2": 1.0, "3": 1.5},
        )
        agent_state_2 = AgentState(
            agent_address_2,
            {"1": 10},
            {"1": 1.0},
            {"2": 1, "3": 2},
            {"2": 1.0, "3": 1.5},
        )

        self.game._current_agent_states = {
            agent_address_1: agent_state_1,
            agent_address_2: agent_state_2,
        }

        # operation
        with patch.object(Transaction, "has_matching_signatures", return_value=True):
            with patch.object(
                AgentState, "is_consistent_transaction", return_value=False
            ):
                assert self.game.is_transaction_valid(tx) is False

    def test_settle_transaction_succeeds(self):
        """Test the settle_transaction method of the Game class which succeeds."""
        # setup
        agent_address_1 = "agent_address_1"
        agent_address_2 = "agent_address_2"

        tx = Transaction(
            "ethereum",
            agent_address_1,
            agent_address_2,
            {"1": 10},
            {"2": -1, "3": 0},
            True,
            "some_nonce",
            {"1": 1},
            "some_sender_signature",
            "some_counterparty_signature",
        )
        agent_state_1 = AgentState(
            agent_address_1,
            {"1": 10},
            {"1": 1.0},
            {"2": 1, "3": 2},
            {"2": 1.0, "3": 1.5},
        )
        agent_state_2 = AgentState(
            agent_address_2,
            {"1": 10},
            {"1": 1.0},
            {"2": 1, "3": 2},
            {"2": 1.0, "3": 1.5},
        )

        self.game._current_agent_states = {
            agent_address_1: agent_state_1,
            agent_address_2: agent_state_2,
        }

        expected_agent_state_1 = AgentState(
            agent_address_1,
            {"1": 20},
            {"1": 1.0},
            {"2": 0, "3": 2},
            {"2": 1.0, "3": 1.5},
        )
        expected_agent_state_2 = AgentState(
            agent_address_2,
            {"1": 0},
            {"1": 1.0},
            {"2": 2, "3": 2},
            {"2": 1.0, "3": 1.5},
        )

        # before
        assert self.game._current_agent_states[agent_address_1] == agent_state_1
        assert self.game._current_agent_states[agent_address_2] == agent_state_2

        # operation
        with patch.object(Transaction, "has_matching_signatures", return_value=True):
            self.game.settle_transaction(tx)

        # after
        assert (
            self.game._current_agent_states[agent_address_1] == expected_agent_state_1
        )
        assert (
            self.game._current_agent_states[agent_address_2] == expected_agent_state_2
        )

    def test_settle_transaction_fails_current_agent_states_is_none(self):
        """Test the settle_transaction method of the Game class which fails because current_agent_states is None."""
        # before
        agent_address_1 = "agent_address_1"
        agent_address_2 = "agent_address_2"

        tx = Transaction(
            "ethereum",
            agent_address_1,
            agent_address_2,
            {"1": 10},
            {"2": 5, "5": 10},
            True,
            "some_nonce",
            {"1": 1},
            "some_sender_signature",
            "some_counterparty_signature",
        )

        # operation
        with pytest.raises(
            AEAEnforceError, match="Call create before calling current_agent_states."
        ):
            assert self.game.settle_transaction(tx)

    def test_settle_transaction_fails_tx_invalid(self):
        """Test the settle_transaction method of the Game class which fails because transaction is invalid."""
        # before
        agent_address_1 = "agent_address_1"
        agent_address_2 = "agent_address_2"

        tx = Transaction(
            "ethereum",
            agent_address_1,
            agent_address_2,
            {"1": 10},
            {"2": 5, "5": 10},
            True,
            "some_nonce",
            {"1": 1},
            "some_sender_signature",
            "some_counterparty_signature",
        )
        self.game._current_agent_states = "some_current_agent_states"

        # operation
        with patch.object(self.game, "is_transaction_valid", return_value=False):
            with pytest.raises(AEAEnforceError, match="Transaction is not valid."):
                assert self.game.settle_transaction(tx)

    def test_get_location_description(self):
        """Test the get_location_description method of the Game class."""
        description = self.game.get_location_description()

        assert type(description) == Description
        assert description.data_model is AGENT_LOCATION_MODEL
        assert description.values.get("location", "") == Location(
            longitude=0.1270, latitude=51.5194
        )

    def test_get_register_tac_description(self):
        """Test the get_register_tac_description method of the Game class."""
        description = self.game.get_register_tac_description()

        assert type(description) == Description
        assert description.data_model is AGENT_SET_SERVICE_MODEL
        assert description.values.get("key", "") == "tac"
        assert description.values.get("value", "") == "v1"

    def test_get_unregister_tac_description(self):
        """Test the get_unregister_tac_description method of the Game class."""
        description = self.game.get_unregister_tac_description()

        assert type(description) == Description
        assert description.data_model is AGENT_REMOVE_SERVICE_MODEL
        assert description.values.get("key", "") == "tac"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains tests for aea exceptions."""

import pytest

from aea.exceptions import AEAEnforceError, _StopRuntime, enforce, parse_exception


def test_enforce_no_exception():
    """Test enforce does not throw exception if condition is True."""
    enforce(True, "Error message")


def test_enforce_exception():
    """Test enforce does throw exception if condition is False."""
    error_msg = "Error message"
    with pytest.raises(AEAEnforceError, match=error_msg):
        enforce(False, error_msg)


def test_stop_runtime():
    """Test thes stop runtime exception."""
    test = "test string"
    e = _StopRuntime(test)
    assert e.reraise == test


def test_parse_exception():
    """Test parse exception."""

    def exception_raise():
        """A function that raises an exception."""
        raise ValueError("expected")

    try:
        exception_raise()
    except Exception as e:
        out = parse_exception(e)

    expected = [
        "Traceback (most recent call last):\n\n",
        'test_exceptions.py", line ',
        "in exception_raise\n",
        'raise ValueError("expected")\n\nValueError: expected\n',
    ]
    assert all([string in out for string in expected])



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the base module."""
import asyncio
import os
import unittest
from pathlib import Path
from typing import cast
from unittest.mock import MagicMock

import pytest

import aea
from aea.configurations.base import (
    ComponentId,
    ComponentType,
    ConnectionConfig,
    PublicId,
)
from aea.configurations.loader import load_component_configuration
from aea.connections.base import Connection, ConnectionStates
from aea.exceptions import AEAComponentLoadException, AEAEnforceError
from aea.mail.base import Envelope

from tests.conftest import CUR_PATH


class TConnection(Connection):
    """Test class for Connection."""

    connection_id = PublicId.from_str("fetchai/some_connection:0.1.0")

    def connect(self, *args, **kwargs):
        """Connect."""
        pass

    def disconnect(self, *args, **kwargs):
        """Disconnect."""
        pass

    def from_config(self, *args, **kwargs):
        """From config."""
        pass

    def receive(self, *args, **kwargs):
        """Receive."""
        pass

    def send(self, *args, **kwargs):
        """Send."""
        pass


class TestConnectionTestCase:
    """Test case for Connection abstract class."""

    TConnection = TConnection

    @pytest.mark.asyncio
    async def test_loop_only_in_running_loop(self):
        """Test loop property positive result."""
        obj = self.TConnection(ConnectionConfig("some_connection", "fetchai", "0.1.0"))
        obj.loop

    def test_loop_fails_on_non_running_loop(self):
        """Test loop property positive result."""
        obj = self.TConnection(ConnectionConfig("some_connection", "fetchai", "0.1.0"))
        with pytest.raises(AEAEnforceError):
            obj.loop

    def test_excluded_protocols_positive(self):
        """Test excluded_protocols property positive result."""
        obj = self.TConnection(ConnectionConfig("some_connection", "fetchai", "0.1.0"))
        obj._excluded_protocols = "excluded_protocols"
        obj.excluded_protocols


def test_loop_property():
    """Test connection's loop property."""
    connection = TConnection(MagicMock(public_id=TConnection.connection_id))
    with unittest.mock.patch.object(aea.connections.base, "enforce"):
        loop = connection.loop
        assert isinstance(loop, asyncio.AbstractEventLoop)


def test_ensure_valid_envelope_for_external_comms_negative_cases():
    """Test the staticmethod '_ensure_valid_envelope_for_external_comms', negative cases."""
    protocol_id = PublicId("author", "name", "0.1.0")
    wrong_sender = wrong_to = "author/name:0.1.0"
    envelope_wrong_sender = Envelope(
        to="to", sender=wrong_sender, protocol_id=protocol_id, message=b""
    )
    with pytest.raises(
        AEAEnforceError,
        match=f"Sender field of envelope is public id, needs to be address. Found={wrong_to}",
    ):
        Connection._ensure_valid_envelope_for_external_comms(envelope_wrong_sender)

    envelope_wrong_sender = Envelope(
        to=wrong_to, sender="sender", protocol_id=protocol_id, message=b""
    )
    with pytest.raises(
        AEAEnforceError,
        match=f"To field of envelope is public id, needs to be address. Found={wrong_sender}",
    ):
        Connection._ensure_valid_envelope_for_external_comms(envelope_wrong_sender)


def test_state():
    """Test connect context of a connection."""
    connection = TConnection(MagicMock(public_id=TConnection.connection_id))
    assert connection.state == ConnectionStates.disconnected

    with connection._connect_context():
        assert connection.state == ConnectionStates.connecting

    assert connection.state == ConnectionStates.connected


def test_from_dir():
    """Test Connection.from_dir"""
    dummy_connection_dir = os.path.join(CUR_PATH, "data", "dummy_connection")
    identity = MagicMock()
    identity.name = "agent_name"
    crypto_store = MagicMock()
    connection = Connection.from_dir(dummy_connection_dir, identity, crypto_store)
    assert isinstance(connection, Connection)
    assert connection.component_id == ComponentId(
        ComponentType.CONNECTION, PublicId("fetchai", "dummy", "0.1.0")
    )


def test_from_config_exception_path():
    """Test Connection.from_config with exception"""
    dummy_connection_dir = os.path.join(CUR_PATH, "data", "dummy_connection")
    configuration = cast(
        ConnectionConfig,
        load_component_configuration(
            ComponentType.CONNECTION, Path(dummy_connection_dir)
        ),
    )
    wrong_dir = os.path.join(CUR_PATH, "data", "wrong_connection")
    configuration.directory = Path(wrong_dir)
    identity = MagicMock()
    identity.name = "agent_name"
    crypto_store = MagicMock()
    with pytest.raises(AEAComponentLoadException, match="Connection module"):
        Connection.from_config(configuration, identity, crypto_store)


def test_from_config_exception_class():
    """Test Connection.from_config with exception"""
    dummy_connection_dir = os.path.join(CUR_PATH, "data", "dummy_connection")
    configuration = cast(
        ConnectionConfig,
        load_component_configuration(
            ComponentType.CONNECTION, Path(dummy_connection_dir)
        ),
    )
    configuration.directory = Path(dummy_connection_dir)
    configuration.class_name = "WrongName"
    identity = MagicMock()
    identity.name = "agent_name"
    crypto_store = MagicMock()
    with pytest.raises(AEAComponentLoadException, match="Connection class"):
        Connection.from_config(configuration, identity, crypto_store)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour classes of the tac control contract skill."""

import datetime
import logging
from pathlib import Path
from typing import cast
from unittest.mock import Mock, patch

from aea.helpers.search.models import Description
from aea.test_tools.test_skill import BaseSkillTestCase, COUNTERPARTY_ADDRESS

from packages.fetchai.protocols.contract_api.message import ContractApiMessage
from packages.fetchai.skills.tac_control_contract.behaviours import (
    LEDGER_API_ADDRESS,
    TacBehaviour,
)
from packages.fetchai.skills.tac_control_contract.dialogues import (
    ContractApiDialogue,
    ContractApiDialogues,
    TacDialogues,
)
from packages.fetchai.skills.tac_control_contract.game import (
    AgentState,
    Configuration,
    Game,
    Phase,
)
from packages.fetchai.skills.tac_control_contract.parameters import Parameters

from tests.conftest import ROOT_DIR


class TestSkillBehaviour(BaseSkillTestCase):
    """Test tac behaviour of tac_control_contract."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "fetchai", "skills", "tac_control_contract"
    )

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.tac_behaviour = cast(TacBehaviour, cls._skill.skill_context.behaviours.tac)
        cls.game = cast(Game, cls._skill.skill_context.game)
        cls.parameters = cast(Parameters, cls._skill.skill_context.parameters)
        cls.tac_dialogues = cast(TacDialogues, cls._skill.skill_context.tac_dialogues)
        cls.contract_api_dialogues = cast(
            ContractApiDialogues, cls._skill.skill_context.contract_api_dialogues
        )
        cls.logger = cls.tac_behaviour.context.logger

        cls.mocked_reg_start_time = cls._time("00:02")
        cls.mocked_reg_end_time = cls._time("00:04")
        cls.mocked_start_time = cls._time("00:06")
        cls.mocked_end_time = cls._time("00:08")

        cls.parameters._registration_start_time = cls.mocked_reg_start_time
        cls.parameters._registration_end_time = cls.mocked_reg_end_time
        cls.parameters._start_time = cls.mocked_start_time
        cls.parameters._end_time = cls.mocked_end_time

        cls.mocked_description = Description({"foo1": 1, "bar1": 2})

        cls.agent_1_address = "agent_address_1"
        cls.agent_1_name = "agent_name_1"
        cls.agent_2_address = "agent_address_2"
        cls.agent_2_name = "agent_name_2"

        cls.amount_by_currency_id = {"1": 10}
        cls.exchange_params_by_currency_id = {"1": 1.0}
        cls.quantities_by_good_id = {"2": 1, "3": 2}
        cls.utility_params_by_good_id = {"2": 1.0, "3": 1.5}

        cls.agent_1_state = AgentState(
            cls.agent_1_address,
            cls.amount_by_currency_id,
            cls.exchange_params_by_currency_id,
            cls.quantities_by_good_id,
            cls.utility_params_by_good_id,
        )
        cls.agent_2_state = AgentState(
            cls.agent_2_address,
            cls.amount_by_currency_id,
            cls.exchange_params_by_currency_id,
            cls.quantities_by_good_id,
            cls.utility_params_by_good_id,
        )

    def test_setup(self):
        """Test the setup method of the tac behaviour."""
        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.tac_behaviour.setup()

        # after
        assert self.game.phase == Phase.CONTRACT_DEPLOYMENT_PROPOSAL

        self.assert_quantity_in_outbox(2)

        # first message is produced in superclass (from tac_control skill) which has its own unit tests
        self.drop_messages_from_outbox(1)

        # _request_contract_deploy_transaction
        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
            to=LEDGER_API_ADDRESS,
            sender=self.skill.skill_context.agent_address,
            ledger_id=self.parameters.ledger_id,
            contract_id=self.parameters.contract_id,
            callable=ContractApiDialogue.Callable.GET_DEPLOY_TRANSACTION.value,
            kwargs=ContractApiMessage.Kwargs(
                {"deployer_address": self.skill.skill_context.agent_address}
            ),
        )
        assert has_attributes, error_str
        assert (
            cast(
                ContractApiDialogue, self.contract_api_dialogues.get_dialogue(message)
            ).terms
            == self.parameters.get_deploy_terms()
        )
        assert (
            cast(
                ContractApiDialogue, self.contract_api_dialogues.get_dialogue(message)
            ).callable
            == ContractApiDialogue.Callable.GET_DEPLOY_TRANSACTION
        )
        mock_logger.assert_any_call(
            logging.INFO, "requesting contract deployment transaction..."
        )

    @staticmethod
    def _time(time: str):
        date_time = "01 01 2020  " + time
        return datetime.datetime.strptime(date_time, "%d %m %Y %H:%M")

    @staticmethod
    def _mock_time(time: str) -> Mock:
        mocked_now_time = TestSkillBehaviour._time(time)

        datetime_mock = Mock(wraps=datetime.datetime)
        datetime_mock.now.return_value = mocked_now_time
        return datetime_mock

    def test_act_i(self):
        """Test the act method of the tac behaviour where phase is contract_deployed and reg_start_time < now < reg_end_time."""
        # setup
        self.game._phase = Phase.CONTRACT_DEPLOYED

        mocked_now = self._mock_time("00:03")

        # operation
        with patch("datetime.datetime", new=mocked_now):
            with patch.object(self.tac_behaviour, "_register_tac") as mock_register_tac:
                with patch.object(self.logger, "log") as mock_logger:
                    self.tac_behaviour.act()

        # after
        assert self.game.phase == Phase.GAME_REGISTRATION

        # _register_tac is a superclass method with its own unit test in test_tac_control
        mock_register_tac.assert_called_once()

        mock_logger.assert_any_call(
            logging.INFO, f"TAC open for registration until: {self.mocked_reg_end_time}"
        )

    def test_act_ii(self):
        """Test the act method of the tac behaviour where phase is game_registration and reg_end_time < now < start_time and nb_agent < min_nb_agents."""
        # setup
        self.game._phase = Phase.GAME_REGISTRATION

        mocked_now = self._mock_time("00:05")

        self.parameters._min_nb_agents = 2
        self.game._registration.register_agent(COUNTERPARTY_ADDRESS, self.agent_1_name)

        # operation
        with patch("datetime.datetime", new=mocked_now):
            with patch.object(self.tac_behaviour, "_cancel_tac") as mock_cancel_tac:
                with patch.object(
                    self.tac_behaviour, "_unregister_tac"
                ) as mock_unregister_tac:
                    with patch.object(self.logger, "log") as mock_logger:
                        self.tac_behaviour.act()

        # after
        mock_logger.assert_any_call(logging.INFO, "closing registration!")
        mock_logger.assert_any_call(
            logging.INFO,
            f"registered agents={1}, minimum agents required={self.parameters._min_nb_agents}",
        )

        # _cancel_tac is a superclass method with its own unit test in test_tac_control
        mock_cancel_tac.assert_called_with(self.game)

        assert self.game.phase == Phase.POST_GAME

        # _unregister_tac is a superclass method with its own unit test in test_tac_control
        mock_unregister_tac.assert_called_once()

        assert self.skill.skill_context.is_active is False

    def test_act_iii(self):
        """Test the act method of the tac behaviour where phase is game_registration and reg_end_time < now < start_time and nb_agent >= min_nb_agents."""
        # setup
        self.game._phase = Phase.GAME_REGISTRATION

        mocked_now = self._mock_time("00:05")

        self.parameters._contract_address = "some_contract_address"

        self.parameters._min_nb_agents = 2
        self.game._registration.register_agent(self.agent_1_address, self.agent_1_name)
        self.game._registration.register_agent(self.agent_2_address, self.agent_2_name)

        # operation
        with patch("datetime.datetime", new=mocked_now):
            with patch.object(
                self.tac_behaviour, "_unregister_tac"
            ) as mock_unregister_tac:
                with patch.object(self.logger, "log") as mock_logger:
                    self.tac_behaviour.act()

        # after
        mock_logger.assert_any_call(logging.INFO, "closing registration!")
        assert self.game.phase == Phase.GAME_SETUP
        assert self.game.conf.contract_address == self.parameters.contract_address

        # _unregister_tac is a superclass method with its own unit test in test_tac_control
        mock_unregister_tac.assert_called_once()

    def test_act_iv(self):
        """Test the act method of the tac behaviour where phase is GAME_SETUP and reg_end_time < now < start_time."""
        # setup
        self.game._phase = Phase.GAME_SETUP

        mocked_now = self._mock_time("00:05")

        self.parameters._min_nb_agents = 2
        self.game._registration.register_agent(self.agent_1_address, self.agent_1_name)
        self.game._registration.register_agent(self.agent_2_address, self.agent_2_name)

        self.game._conf = Configuration(
            "v1",
            1,
            self.game.registration.agent_addr_to_name,
            {"1": "currency_name"},
            {"2": "good_1", "3": "good_2"},
        )

        self.parameters._contract_address = "some_contract_address"

        # operation
        with patch("datetime.datetime", new=mocked_now):
            with patch.object(self.logger, "log") as mock_logger:
                self.tac_behaviour.act()

        # after
        assert self.game.phase == Phase.TOKENS_CREATION_PROPOSAL

        # _request_create_items_transaction
        self.assert_quantity_in_outbox(1)
        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            to=LEDGER_API_ADDRESS,
            sender=self.skill.skill_context.agent_address,
            ledger_id=self.parameters.ledger_id,
            contract_id=self.parameters.contract_id,
            contract_address=self.parameters.contract_address,
            callable=ContractApiDialogue.Callable.GET_CREATE_BATCH_TRANSACTION.value,
            kwargs=ContractApiMessage.Kwargs(
                {
                    "deployer_address": self.skill.skill_context.agent_address,
                    "token_ids": [2, 3, 1],
                }
            ),
        )
        assert has_attributes, error_str
        assert (
            cast(
                ContractApiDialogue, self.contract_api_dialogues.get_dialogue(message)
            ).terms
            == self.parameters.get_create_token_terms()
        )
        assert (
            cast(
                ContractApiDialogue, self.contract_api_dialogues.get_dialogue(message)
            ).callable
            == ContractApiDialogue.Callable.GET_CREATE_BATCH_TRANSACTION
        )
        mock_logger.assert_any_call(
            logging.INFO, "requesting create items transaction..."
        )

    def test_act_v(self):
        """Test the act method of the tac behaviour where phase is TOKENS_CREATED."""
        # setup
        self.game._phase = Phase.TOKENS_CREATED

        self.parameters._min_nb_agents = 2
        self.game._registration.register_agent(self.agent_1_address, self.agent_1_name)
        self.game._registration.register_agent(self.agent_2_address, self.agent_2_name)

        self.game._initial_agent_states = {
            self.agent_1_address: self.agent_1_state,
            self.agent_2_address: self.agent_2_state,
        }

        self.parameters._contract_address = "some_contract_address"

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.tac_behaviour.act()

        # after
        assert self.game.phase == Phase.TOKENS_MINTING_PROPOSAL

        # _request_mint_items_transaction
        assert self.game.is_allowed_to_mint is False

        mock_logger.assert_any_call(
            logging.INFO,
            f"requesting mint_items transactions for agent={self.agent_1_name}.",
        )

        self.assert_quantity_in_outbox(1)
        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            to=LEDGER_API_ADDRESS,
            sender=self.skill.skill_context.agent_address,
            ledger_id=self.parameters.ledger_id,
            contract_id=self.parameters.contract_id,
            contract_address=self.parameters.contract_address,
            callable=ContractApiDialogue.Callable.GET_MINT_BATCH_TRANSACTION.value,
            kwargs=ContractApiMessage.Kwargs(
                {
                    "deployer_address": self.skill.skill_context.agent_address,
                    "recipient_address": self.skill.skill_context.agent_address,
                    "token_ids": [2, 3, 1],
                    "mint_quantities": [1, 2, 10],
                }
            ),
        )
        assert has_attributes, error_str

        assert (
            cast(
                ContractApiDialogue, self.contract_api_dialogues.get_dialogue(message)
            ).terms
            == self.parameters.get_mint_token_terms()
        )
        assert (
            cast(
                ContractApiDialogue, self.contract_api_dialogues.get_dialogue(message)
            ).callable
            == ContractApiDialogue.Callable.GET_MINT_BATCH_TRANSACTION
        )

    def test_request_mint_items_transaction_not_allowed_to_mint(self):
        """Test the _request_mint_items_transaction method of the tac behaviour where is_allowed_to_mint is False."""
        # setup
        self.game._phase = Phase.TOKENS_CREATED
        self.game.is_allowed_to_mint = False

        # operation
        self.tac_behaviour.act()

        # after
        assert self.game.phase == Phase.TOKENS_MINTING_PROPOSAL

        # _request_mint_items_transaction
        self.assert_quantity_in_outbox(0)

    def test_request_mint_items_transaction_agent_state_is_none(self):
        """Test the _request_mint_items_transaction method of the tac behaviour where agent_state is None."""
        # setup
        self.game._phase = Phase.TOKENS_CREATED
        self.game._initial_agent_states = {self.agent_1_address: None}

        # operation
        self.tac_behaviour.act()

        # after
        assert self.game.phase == Phase.TOKENS_MINTING_PROPOSAL

        # _request_mint_items_transaction
        assert self.game.is_allowed_to_mint is False
        self.assert_quantity_in_outbox(0)

    def test_act_vi(self):
        """Test the act method of the tac behaviour where phase is TOKENS_MINTING_PROPOSAL."""
        # setup
        self.game._phase = Phase.TOKENS_MINTING_PROPOSAL

        self.parameters._min_nb_agents = 2
        self.game._registration.register_agent(self.agent_1_address, self.agent_1_name)
        self.game._registration.register_agent(self.agent_2_address, self.agent_2_name)

        self.game._initial_agent_states = {
            self.agent_1_address: self.agent_1_state,
            self.agent_2_address: self.agent_2_state,
        }

        self.parameters._contract_address = "some_contract_address"

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.tac_behaviour.act()

        # after
        assert self.game.phase == Phase.TOKENS_MINTING_PROPOSAL

        # _request_mint_items_transaction
        assert self.game.is_allowed_to_mint is False

        mock_logger.assert_any_call(
            logging.INFO,
            f"requesting mint_items transactions for agent={self.agent_1_name}.",
        )

        self.assert_quantity_in_outbox(1)
        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            to=LEDGER_API_ADDRESS,
            sender=self.skill.skill_context.agent_address,
            ledger_id=self.parameters.ledger_id,
            contract_id=self.parameters.contract_id,
            contract_address=self.parameters.contract_address,
            callable=ContractApiDialogue.Callable.GET_MINT_BATCH_TRANSACTION.value,
            kwargs=ContractApiMessage.Kwargs(
                {
                    "deployer_address": self.skill.skill_context.agent_address,
                    "recipient_address": self.skill.skill_context.agent_address,
                    "token_ids": [2, 3, 1],
                    "mint_quantities": [1, 2, 10],
                }
            ),
        )
        assert has_attributes, error_str

        assert (
            cast(
                ContractApiDialogue, self.contract_api_dialogues.get_dialogue(message)
            ).terms
            == self.parameters.get_mint_token_terms()
        )
        assert (
            cast(
                ContractApiDialogue, self.contract_api_dialogues.get_dialogue(message)
            ).callable
            == ContractApiDialogue.Callable.GET_MINT_BATCH_TRANSACTION
        )

    def test_act_vii(self):
        """Test the act method of the tac behaviour where phase is TOKENS_MINTED and start_time < now < end_time."""
        # setup
        self.game._phase = Phase.TOKENS_MINTED

        mocked_now = self._mock_time("00:07")

        # operation
        with patch("datetime.datetime", new=mocked_now):
            with patch.object(self.tac_behaviour, "_start_tac") as mock_start_tac:
                with patch.object(self.logger, "log"):
                    self.tac_behaviour.act()

        # after
        assert self.game.phase == Phase.GAME

        # _start_tac is a superclass method with its own unit test in test_tac_control
        mock_start_tac.assert_called_with(self.game)

    def test_act_viii(self):
        """Test the act method of the tac behaviour where phase is GAME and end_time < now."""
        # setup
        self.game._phase = Phase.GAME

        mocked_now = self._mock_time("00:09")

        # operation
        with patch("datetime.datetime", new=mocked_now):
            with patch.object(self.tac_behaviour, "_cancel_tac") as mock_cancel_tac:
                with patch.object(self.logger, "log"):
                    self.tac_behaviour.act()

        # after
        assert self.game.phase == Phase.POST_GAME

        # _cancel_tac is a superclass method with its own unit test in test_tac_control
        mock_cancel_tac.assert_called_with(self.game)

        assert self.skill.skill_context.is_active is False



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for commands in aea.cli.eject module."""

import os
from pathlib import Path
from unittest import mock

import click
import pytest

from aea.cli.utils.config import get_or_create_cli_config
from aea.configurations.base import ComponentType, DEFAULT_VERSION, PublicId
from aea.configurations.loader import load_component_configuration
from aea.test_tools.test_cases import AEATestCaseEmpty, AEATestCaseMany

from packages.fetchai.connections.gym.connection import (
    PUBLIC_ID as GYM_CONNECTION_PUBLIC_ID,
)
from packages.fetchai.contracts.erc1155.contract import PUBLIC_ID as ERC1155_PUBLIC_ID
from packages.fetchai.protocols.default import DefaultMessage
from packages.fetchai.protocols.gym.message import GymMessage
from packages.fetchai.skills.error import PUBLIC_ID as ERROR_PUBLIC_ID
from packages.fetchai.skills.gym import PUBLIC_ID as GYM_SKILL_PUBLIC_ID


class TestEjectCommands(AEATestCaseMany):
    """End-to-end test case for CLI eject commands."""

    def test_eject_commands_positive(self):
        """Test eject commands for positive result."""
        agent_name = "test_aea"
        self.create_agents(agent_name)

        self.set_agent_context(agent_name)
        cwd = os.path.join(self.t, agent_name)
        self.add_item("connection", str(GYM_CONNECTION_PUBLIC_ID))
        self.add_item("skill", str(GYM_SKILL_PUBLIC_ID))
        self.add_item("contract", str(ERC1155_PUBLIC_ID))

        # the order must be kept as is, because of recursive ejects
        self.eject_item("skill", str(GYM_SKILL_PUBLIC_ID))
        assert "gym" not in os.listdir(
            (os.path.join(cwd, "vendor", "fetchai", "skills"))
        )
        assert "gym" in os.listdir((os.path.join(cwd, "skills")))
        self.eject_item("connection", str(GYM_CONNECTION_PUBLIC_ID))
        assert "gym" not in os.listdir(
            (os.path.join(cwd, "vendor", "fetchai", "connections"))
        )
        assert "gym" in os.listdir((os.path.join(cwd, "connections")))

        self.eject_item("protocol", str(GymMessage.protocol_id))
        assert "gym" not in os.listdir(
            (os.path.join(cwd, "vendor", "fetchai", "protocols"))
        )
        assert "gym" in os.listdir((os.path.join(cwd, "protocols")))

        self.eject_item("contract", str(ERC1155_PUBLIC_ID))
        assert "erc1155" not in os.listdir(
            (os.path.join(cwd, "vendor", "fetchai", "contracts"))
        )
        assert "erc1155" in os.listdir((os.path.join(cwd, "contracts")))


class TestRecursiveEject(AEATestCaseEmpty):
    """Test that eject is recursive."""

    def test_recursive_eject_commands_positive(self):
        """Test eject commands for positive result."""
        agent_name = "test_aea"
        self.create_agents(agent_name)

        self.set_agent_context(agent_name)
        cwd = os.path.join(self.t, agent_name)
        self.add_item("connection", str(GYM_CONNECTION_PUBLIC_ID))
        self.add_item("skill", str(GYM_SKILL_PUBLIC_ID))
        self.add_item("contract", str(ERC1155_PUBLIC_ID))

        # ejecting the gym protocol will cause the ejection of
        # all the other packages that depend on it,
        # that is, gym connection and gym skill.
        self.eject_item("protocol", str(GymMessage.protocol_id))
        assert "gym" not in os.listdir(
            (os.path.join(cwd, "vendor", "fetchai", "protocols"))
        )
        assert "gym" in os.listdir((os.path.join(cwd, "protocols")))
        assert "gym" not in os.listdir(
            (os.path.join(cwd, "vendor", "fetchai", "connections"))
        )
        assert "gym" in os.listdir((os.path.join(cwd, "connections")))
        assert "gym" not in os.listdir(
            (os.path.join(cwd, "vendor", "fetchai", "skills"))
        )
        assert "gym" in os.listdir((os.path.join(cwd, "skills")))


class TestRecursiveEjectIsAborted(AEATestCaseEmpty):
    """Test that recursive eject is aborted in non-quiet mode."""

    @mock.patch("click.confirm", return_value=False)
    def test_recursive_eject_commands_non_quiet_negative(self, *_mocks):
        """Test eject command for negative result in interactive mode."""
        agent_name = "test_aea"
        self.create_agents(agent_name)

        self.set_agent_context(agent_name)
        cwd = os.path.join(self.t, agent_name)
        self.add_item("connection", str(GYM_CONNECTION_PUBLIC_ID))
        self.add_item("skill", str(GYM_SKILL_PUBLIC_ID))
        self.add_item("contract", str(ERC1155_PUBLIC_ID))

        self.run_cli_command(
            "eject", "protocol", str(GymMessage.protocol_id), cwd=self._get_cwd()
        )
        # assert packages not ejected
        assert "gym" in os.listdir(
            (os.path.join(cwd, "vendor", "fetchai", "protocols"))
        )
        assert "gym" not in os.listdir((os.path.join(cwd, "protocols")))
        assert "gym" in os.listdir(
            (os.path.join(cwd, "vendor", "fetchai", "connections"))
        )
        assert "gym" not in os.listdir((os.path.join(cwd, "connections")))
        assert "gym" in os.listdir((os.path.join(cwd, "vendor", "fetchai", "skills")))
        assert "gym" not in os.listdir((os.path.join(cwd, "skills")))


class BaseTestEjectCommand(AEATestCaseEmpty):
    """Replace CLI author with a known author."""

    EXPECTED_AUTHOR = ""

    @classmethod
    def setup_class(cls):
        """Set up the class."""
        super().setup_class()
        config = get_or_create_cli_config()
        cls.EXPECTED_AUTHOR = config.get("author", "")


class TestEjectCommandCliConfigNotAvailable(AEATestCaseEmpty):
    """Test that 'aea eject' cannot be run if CLI configuration not provided."""

    IS_EMPTY = True

    @classmethod
    def setup_class(cls):
        """Set up the class."""
        super().setup_class()
        cls.add_item("protocol", str(DefaultMessage.protocol_id))

    @mock.patch("aea.cli.utils.config.get_or_create_cli_config", return_value={})
    def test_error(self, *_mocks):
        """Test that without CLI configuration, 'aea eject' won't work."""
        with pytest.raises(
            click.ClickException,
            match="The AEA configurations are not initialized. Use `aea init` before continuing.",
        ):
            self.invoke(
                "eject", "--quiet", "protocol", str(DefaultMessage.protocol_id),
            )


class TestEjectCommandReplacesReferences(BaseTestEjectCommand):
    """Test that eject command replaces the right references to the new package."""

    IS_EMPTY = True

    @classmethod
    def setup_class(cls):
        """Set up the class."""
        super().setup_class()
        cls.add_item("protocol", str(DefaultMessage.protocol_id))
        cls.eject_item("protocol", str(DefaultMessage.protocol_id))

    def test_username_is_correct(self):
        """Test that the author name in the ejected component configuration is updated correctly."""
        package_path = Path(
            self.current_agent_context, "protocols", DefaultMessage.protocol_id.name
        )
        assert (
            package_path.exists()
        ), f"Expected ejected package in '{package_path}', but not found."
        component_configuration = load_component_configuration(
            ComponentType.PROTOCOL, package_path
        )
        assert component_configuration.author == self.EXPECTED_AUTHOR
        assert component_configuration.name == DefaultMessage.protocol_id.name
        assert component_configuration.version == DEFAULT_VERSION

    def test_aea_config_references_updated_correctly(self):
        """Test that the references in the AEA configuration is updated correctly."""
        agent_config = self.load_agent_config(self.agent_name)
        assert agent_config.protocols == {
            PublicId(
                self.EXPECTED_AUTHOR, DefaultMessage.protocol_id.name, DEFAULT_VERSION
            )
        }


class TestEjectCommandReplacesCustomConfigurationReference(BaseTestEjectCommand):
    """Test that eject command replaces references in AEA configuration."""

    IS_EMPTY = True

    @classmethod
    def setup_class(cls):
        """Set up the class."""
        super().setup_class()
        cls.add_item("skill", str(ERROR_PUBLIC_ID))
        # add a custom configuration to the error skill
        cls.run_cli_command(
            "--skip-consistency-check",
            "config",
            "set",
            "vendor.fetchai.skills.error.is_abstract",
            "--type",
            "bool",
            "true",
            cwd=cls._get_cwd(),
        )

        cls.eject_item("skill", str(ERROR_PUBLIC_ID))

    def test_username_is_correct(self):
        """Test that the author name in the ejected component configuration is updated correctly."""
        package_path = Path(self.current_agent_context, "skills", ERROR_PUBLIC_ID.name)
        assert (
            package_path.exists()
        ), f"Expected ejected package in '{package_path}', but not found."
        component_configuration = load_component_configuration(
            ComponentType.SKILL, package_path
        )
        assert component_configuration.author == self.EXPECTED_AUTHOR

    def test_aea_config_references_updated_correctly(self):
        """Test that the references in the AEA configuration is updated correctly."""
        agent_config = self.load_agent_config(self.agent_name)
        assert agent_config.skills == {
            PublicId(self.EXPECTED_AUTHOR, ERROR_PUBLIC_ID.name, DEFAULT_VERSION)
        }


class TestEjectWithLatest(AEATestCaseEmpty):
    """Test the eject command when a public id 'latest' is provided."""

    def test_command(self):
        """Run the test."""
        latest_public_id = ERROR_PUBLIC_ID.to_latest()
        self.eject_item("skill", str(latest_public_id))
        cwd = os.path.join(self.t, self.agent_name)
        # assert packages ejected
        assert "error" not in os.listdir(
            (os.path.join(cwd, "vendor", "fetchai", "skills"))
        )
        assert "error" in os.listdir((os.path.join(cwd, "skills")))


class TestEjectWithSymlink(AEATestCaseEmpty):
    """Test the eject command with symlinks flag."""

    def test_command(self):
        """Run the test."""
        latest_public_id = ERROR_PUBLIC_ID.to_latest()
        self.run_cli_command(
            "eject",
            "--with-symlinks",
            "skill",
            str(latest_public_id),
            cwd=self._get_cwd(),
        )
        cwd = os.path.join(self.t, self.agent_name)
        # assert packages ejected
        assert "error" not in os.listdir(
            (os.path.join(cwd, "vendor", "fetchai", "skills"))
        )
        assert "error" in os.listdir((os.path.join(cwd, "skills")))
        assert "error" in os.listdir(
            (os.path.join(cwd, "vendor", self.author, "skills"))
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Temporary error handler."""
from logging import Logger

from aea.mail.base import Envelope


class ErrorHandler:
    """Error handler class for handling problematic envelopes."""

    unsupported_protocol_count = 0
    unsupported_skill_count = 0
    decoding_error_count = 0

    @classmethod
    def send_unsupported_protocol(cls, envelope: Envelope, logger: Logger) -> None:
        """
        Handle the received envelope in case the protocol is not supported.

        :param envelope: the envelope
        :return: None
        """
        cls.unsupported_protocol_count += 1
        logger.warning(
            f"Unsupported protocol: {envelope.protocol_id}. You might want to add a handler for this protocol. Sender={envelope.sender}, to={envelope.sender}."
        )

    @classmethod
    def send_decoding_error(cls, envelope: Envelope, logger: Logger) -> None:
        """
        Handle a decoding error.

        :param envelope: the envelope
        :return: None
        """
        cls.decoding_error_count += 1
        logger.warning(
            f"Decoding error for envelope: {envelope}. Protocol_id='{envelope.protocol_id}' and message are inconsistent. Sender={envelope.sender}, to={envelope.sender}."
        )

    @classmethod
    def send_unsupported_skill(cls, envelope: Envelope, logger: Logger) -> None:
        """
        Handle the received envelope in case the skill is not supported.

        :param envelope: the envelope
        :return: None
        """
        cls.unsupported_skill_count += 1
        if envelope.skill_id is None:
            logger.warning(
                f"Cannot handle envelope: no active handler registered for the protocol_id='{envelope.protocol_id}'. Sender={envelope.sender}, to={envelope.sender}."
            )
        else:
            logger.warning(
                f"Cannot handle envelope: no active handler registered for the protocol_id='{envelope.protocol_id}' and skill_id='{envelope.skill_id}'. Sender={envelope.sender}, to={envelope.sender}."
            )
